from typing import Any

from torch import Tensor
from torch.nn import Linear as TorchLinear
from torch.nn import Module
from torch.nn.utils import parametrize as P


class Linear(TorchLinear):
    """A linear layer.
    The weights and bias are fake quantized during initialization.
    Make sure that math_ops is a module where all needed tensors are part of it,
    so they can be moved to the same device.
    Make sure that weight_quantization and bias_quantization are modules that implement the forward function.
    If you want to quantize during initialization or only apply quantized updates make sure to use a quantized optimizer
    and implement the right_inverse method for your module.
    """

    def __init__(
        self,
        math_ops: Module,
        in_features: int,
        out_features: int,
        weight_quantization: Module,
        bias: bool,
        bias_quantization: Module = None,
        device: Any = None,
        dtype: Any = None,
    ) -> None:
        if bias ^ isinstance(bias_quantization, Module):
            raise Exception(
                f"if bias is True, bias_quantization can needs be set. "
                f"If not it is not allowed to be set."
                f"You have choosen {bias=} and {bias_quantization=}."
            )

        super().__init__(
            in_features=in_features,
            out_features=out_features,
            bias=bias,
            device=device,
            dtype=dtype,
        )
        P.register_parametrization(self, "weight", weight_quantization)
        if bias:
            P.register_parametrization(self, "bias", bias_quantization)
        self.add_module("math_ops", math_ops)

    def forward(self, x: Tensor) -> Tensor:
        # return super().forward(x)
        return self.math_ops(super().forward(x))



import pytest

import elasticai.creator.nn.fixed_point as nn_creator
from elasticai.creator.arithmetic import (
    FxpArithmetic,
    FxpParams,
)
from elasticai.creator.nn import Sequential
from tests.integration_tests.nn.fixed_point.sequential_routine import (
    routine_testing_sequential_module,
)


@pytest.mark.simulation
@pytest.mark.parametrize(
    "total_bits, frac_bits, features_in, features_out",
    [
        (6, 4, 20, 12),
        (8, 6, 20, 8),
        (10, 8, 24, 20),
    ],
)
def test_build_test_batchnorm_linear(
    total_bits: int,
    frac_bits: int,
    features_in: int,
    features_out: int,
) -> None:
    file_name = (
        f"TestBatchNormLinear_{total_bits}_{frac_bits}_{features_in}x{features_out}"
    )
    fxp = FxpArithmetic(
        FxpParams(total_bits=total_bits, frac_bits=frac_bits, signed=True)
    )

    dut = Sequential(
        nn_creator.BatchNormedLinear(
            in_features=features_in,
            out_features=features_out,
            total_bits=total_bits,
            frac_bits=frac_bits,
        )
    )

    routine_testing_sequential_module(
        dut=dut,
        file_name=file_name,
        fxp=fxp,
        feat_in=features_in,
        check_quant=False,
    )



from .ir2vhdl import (
    Code,
    Edge,
    Implementation,
    Instance,
    InstanceFactory,
    Ir2Vhdl,
    LogicSignal,
    LogicVectorSignal,
    LoweringPass,
    NullDefinedLogicSignal,
    PluginLoader,
    PluginSpec,
    PluginSymbol,
    PortMap,
    Shape,
    ShapeTuple,
    Signal,
    TypeHandlerFn,
    VhdlNode,
    edge,
    type_handler,
    type_handler_iterable,
    vhdl_node,
)
from .vhdl_template import (
    EntityTemplateDirector,
    EntityTemplateParameter,
    ValueTemplateParameter,
)

__all__ = [
    "Ir2Vhdl",
    "Edge",
    "Code",
    "Signal",
    "Instance",
    "Implementation",
    "InstanceFactory",
    "LogicSignal",
    "LogicVectorSignal",
    "NullDefinedLogicSignal",
    "PluginLoader",
    "LoweringPass",
    "PluginSpec",
    "PluginSymbol",
    "PortMap",
    "Shape",
    "ShapeTuple",
    "TypeHandlerFn",
    "VhdlNode",
    "edge",
    "type_handler",
    "type_handler_iterable",
    "vhdl_node",
    "EntityTemplateParameter",
    "ValueTemplateParameter",
    "EntityTemplateDirector",
]



from .src import grouped_filter
from .src.filter_params import FilterParameters
from .src.index_generators import GroupedFilterIndexGenerator

__all__ = ["FilterParameters", "grouped_filter", "GroupedFilterIndexGenerator"]



import warnings
from abc import abstractmethod
from collections.abc import Callable, Iterable, Iterator
from functools import wraps
from typing import Generic, ParamSpec, Protocol, TypeVar

from elasticai.creator.function_utils import FunctionDecoratorDescriptor
from elasticai.creator.function_utils import KeyedFunctionDispatcher as _Registry


class Lowerable(Protocol):
    @property
    @abstractmethod
    def type(self) -> str: ...


Tin = TypeVar("Tin", bound="Lowerable")
Tout = TypeVar("Tout")


class LoweringPass(Generic[Tin, Tout]):
    register: FunctionDecoratorDescriptor[Tin, Tout] = FunctionDecoratorDescriptor()
    register_override: FunctionDecoratorDescriptor[Tin, Tout] = (
        FunctionDecoratorDescriptor()
    )
    register_iterable: FunctionDecoratorDescriptor[Tin, Iterable[Tout]] = (
        FunctionDecoratorDescriptor()
    )
    register_iterable_override: FunctionDecoratorDescriptor[Tin, Iterable[Tout]] = (
        FunctionDecoratorDescriptor()
    )

    def __init__(self) -> None:
        def key_lookup_fn(x: Tin) -> str:
            return x.type

        self._fns: _Registry[Tin, Iterable[Tout]] = _Registry(key_lookup_fn)

    def _register_callback(self, name: str, fn: Callable[[Tin], Tout]):
        self._check_for_redefinition(name)
        wrapper = return_as_iterable(fn)
        self._fns.register(name)(wrapper)

    def _register_override_callback(self, name: str, fn: Callable[[Tin], Tout]):
        self._check_for_override(name)
        wrapper = return_as_iterable(fn)
        self._fns.register(name)(wrapper)

    def _register_iterable_callback(
        self, name: str, fn: Callable[[Tin], Iterable[Tout]]
    ):
        self._check_for_redefinition(name)
        self._fns.register(name)(fn)

    def _register_iterable_override_callback(
        self, name: str, fn: Callable[[Tin], Iterable[Tout]]
    ):
        self._check_for_override(name)
        self._fns.register(name)(fn)

    def __call__(self, args: Iterable[Tin]) -> Iterator[Tout]:
        for arg in args:
            yield from self._fns(arg)

    def _check_for_redefinition(self, arg):
        if arg in self._fns:
            raise ValueError(f"function for {arg} already defined in lowering pass")

    def _check_for_override(self, arg):
        if arg not in self._fns:
            warnings.warn(
                "expected to override registered function for {}, but no function for that type was defined".format(
                    arg
                ),
                stacklevel=3,
            )


P = ParamSpec("P")


def return_as_iterable(fn: Callable[P, Tout]) -> Callable[P, Iterable[Tout]]:
    @wraps(fn)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> Iterable[Tout]:
        yield fn(*args, **kwargs)

    return wrapper



from itertools import chain
from unittest import TestCase

from elasticai.creator_plugins.grouped_filter.src.index_generators import (
    generate_deinterleaved_indices,
    sliding_window,
)


class InterleaveChannelsAfterGroupwiseOrdering(TestCase):
    def select(self, filtered, indices):
        filtered = tuple(filtered)
        for i in indices:
            yield filtered[i]

    def test_interleave_3bit_output_operations(self):
        # running imaginary filters with kernel size 2 and 3 outputs, two filters per group produces
        expected_indices = [0, 1, 2, 6, 7, 8, 3, 4, 5, 9, 10, 11]
        self.assertEqual(
            expected_indices, list(generate_deinterleaved_indices(12, 2, 3))
        )

    def test_interleave_2bit_output_operations(self):
        expected = tuple(range(8))
        groupwise_output = tuple(chain((0, 1, 4, 5), (2, 3, 6, 7)))
        interleaving_indices = generate_deinterleaved_indices(8, 2, 2)

        self.assertEqual(
            expected,
            tuple(self.select(filtered=groupwise_output, indices=interleaving_indices)),
        )

    def test_interleave_2bit_output_operation_with_length_16_and_8_channels(self):
        expected = tuple(range(16))
        groupwise_output = tuple(
            chain((0, 1, 2, 3, 8, 9, 10, 11), (4, 5, 6, 7, 12, 13, 14, 15))
        )
        interleaving_indices = tuple(generate_deinterleaved_indices(16, 2, 4))
        interleaved = tuple(self.select(groupwise_output, interleaving_indices))
        self.assertEqual(expected, tuple(interleaved))

    def prepare(self, filtered, channels, channel_size):
        filtered = tuple(chain(filtered))
        length = len(filtered)
        indices = tuple(generate_deinterleaved_indices(length, channels, channel_size))
        return tuple(range(length)), tuple(
            self.select(
                filtered,
                indices,
            )
        )

    def test_interleave_2bit_output_with_length_16_and_4_channels(self):
        tuple(chain((0, 1, 4, 5, 8, 9, 12, 13), (2, 3, 6, 7, 10, 11, 14, 15)))
        expected_indices = (0, 1, 8, 9, 2, 3, 10, 11, 4, 5, 12, 13, 6, 7, 14, 15)
        self.assertEqual(
            expected_indices, tuple(generate_deinterleaved_indices(16, 2, 2))
        )

    def test_interleave_2bit_output_with_length_12_and_3_channels(self):
        tuple(chain((0, 1, 6, 7), (2, 3, 8, 9), (4, 5, 10, 11)))
        expected_indices = (0, 1, 4, 5, 8, 9, 2, 3, 6, 7, 10, 11)
        self.assertEqual(
            expected_indices, tuple(generate_deinterleaved_indices(12, 3, 2))
        )

    def test_interleave_3bit_output_with_length_18_and_3_channels(self):
        tuple(chain((0, 1, 2, 9, 10, 11), (3, 4, 5, 12, 13, 14), (6, 7, 8, 15, 16, 17)))
        expected_indices = (
            0,
            1,
            2,
            6,
            7,
            8,
            12,
            13,
            14,
            3,
            4,
            5,
            9,
            10,
            11,
            15,
            16,
            17,
        )
        self.assertEqual(
            expected_indices, tuple(generate_deinterleaved_indices(18, 3, 3))
        )

    def test_interleave_3bit_output_with_length_27_and_3_channels(self):
        groupwise_output = tuple(
            chain(
                (0, 1, 2, 9, 10, 11, 18, 19, 20),
                (3, 4, 5, 12, 13, 14, 21, 22, 23),
                (6, 7, 8, 15, 16, 17, 24, 25, 26),
            )
        )

        self.assertEqual(*self.prepare(groupwise_output, 3, 3))

    def test_interleave_3bit_output_with_length_24_and_2_channels(self):
        groupwise_output = tuple(
            chain(
                (0, 1, 2, 12, 13, 14),
                (3, 4, 5, 15, 16, 17),
                (6, 7, 8, 18, 19, 20),
                (9, 10, 11, 21, 22, 23),
            )
        )
        self.assertEqual(*self.prepare(groupwise_output, 4, 3))


class SlidingWindowIndicesTest(TestCase):
    def get_as_tuple(self, steps, size, stride):
        return tuple(tuple(x) for x in sliding_window(steps, size, stride))

    def test_two_steps(self):
        expected = ((0, 1, 2), (1, 2, 3))
        self.assertEqual(expected, self.get_as_tuple(2, 3, 1))

    def test_stride_2_skips_one_index(self):
        expected = ((0, 1, 2), (2, 3, 4))
        self.assertEqual(expected, self.get_as_tuple(2, 3, 2))

    def test_size_2_stride_3_misses_one_index(self):
        expected = ((0, 1), (3, 4))
        self.assertEqual(expected, self.get_as_tuple(2, 2, 3))



from collections.abc import Callable

import pytest

import elasticai.creator.graph as gr
from elasticai.creator.graph.base_graph import BaseGraph
from elasticai.creator.graph.graph import Graph
from elasticai.creator.graph.vf2 import find_all_matches, match
from elasticai.creator.graph.vf2.matching import MatchError


@pytest.fixture
def create_graph():
    def create():
        return gr.BaseGraph()

    return create


def add_path(graph: gr.Graph, path: str):
    node_path = path.split("->")
    for a, b in zip(node_path[:-1], node_path[1:]):
        graph.add_edge(a, b)


def test_match_single_edge(create_graph: Callable[[], BaseGraph]):
    p = create_graph()
    add_path(p, "a->b")
    g = create_graph()
    add_path(g, "0->1")
    assert match(p, g) == {"a": "0", "b": "1"}


def test_match_single_edge_two_times(create_graph: Callable[[], BaseGraph]):
    p = create_graph()
    add_path(p, "a->b")
    g = create_graph()
    add_path(g, "0->1->2->3")
    assert find_all_matches(p, g) == [
        {"a": "0", "b": "1"},
        {"a": "1", "b": "2"},
        {"a": "2", "b": "3"},
    ]


def test_graph_matches_itself(create_graph: Callable[[], Graph]):
    g = create_graph()
    add_path(g, "1->2")
    add_path(g, "1->3")
    print(g.successors)

    match(g, g)


def test_double_edge(create_graph: Callable[[], BaseGraph]):
    p = create_graph()
    add_path(p, "a->b->c")
    g = create_graph()
    add_path(g, "0->1->2")
    assert match(p, g) == {"a": "0", "b": "1", "c": "2"}


def test_double_out_edge(create_graph: Callable[[], BaseGraph]):
    p = create_graph()
    add_path(p, "a->b")
    g = create_graph()
    for path in ["0->1", "0->2"]:
        add_path(g, path)
    assert match(p, g) in [
        {"a": "0", "b": "1"},
        {"a": "0", "b": "2"},
    ]


def test_donot_match_partially_with_single_edge(create_graph: Callable[[], BaseGraph]):
    p = create_graph()
    add_path(p, "a->b->c")
    g = create_graph()
    add_path(g, "0->1")
    pytest.raises(MatchError, match, p, g)


def test_donot_match_partially_four_to_three_edges(
    create_graph: Callable[[], BaseGraph],
):
    p = create_graph()
    add_path(p, "a->b->c->d->e")
    g = create_graph()
    add_path(g, "0->1->2->3")
    pytest.raises(MatchError, match, p, g)


def test_match_acyclic_pattern_to_cyclic_graph(create_graph: Callable[[], BaseGraph]):
    p = create_graph()
    add_path(p, "a->b")
    g = create_graph()
    add_path(g, "0->1->0")

    assert find_all_matches(p, g) == []


def test_match_cyclic_pattern_to_cyclic_graph(create_graph: Callable[[], BaseGraph]):
    p = create_graph()
    add_path(p, "a->b->a")
    g = create_graph()
    add_path(g, "0->1->0")
    assert match(p, g) in [{"a": "0", "b": "1"}, {"a": "1", "b": "0"}]


def test_node_constraint_can_prevent_first_match(create_graph: Callable[[], BaseGraph]):
    p = create_graph()
    add_path(p, "a->b")
    g = create_graph()
    add_path(g, "0->1->2")
    assert match(p, g, lambda pn, gn: gn in ("1", "2")) == {"a": "1", "b": "2"}


def test_match_diamond_pattern(create_graph: Callable[[], BaseGraph]):
    r"""
    a-b-c-e
     \   /
      d--

    1-2-3-4
     \   /
      6--
    """
    p = create_graph()
    add_path(p, "a->b->c->e")
    add_path(p, "a->d->e")
    g = create_graph()
    add_path(g, "1->2->3->4")
    add_path(g, "1->6->4")
    assert match(p, g) == {"a": "1", "b": "2", "c": "3", "d": "6", "e": "4"}



from pathlib import Path

from elasticai.creator.ir2vhdl.testing import run_vunit_vhdl_testbenches

if __name__ == "__main__":
    run_vunit_vhdl_testbenches(
        [
            "elasticai.creator_plugins.shift_register",
            "elasticai.creator_plugins.counter",
        ],
        Path(__file__).parent,
    )



from elasticai.creator.file_generation.savable import Path
from elasticai.creator.file_generation.template import (
    InProjectTemplate,
    module_to_package,
)
from elasticai.creator.vhdl.design.design import Design


class LSTMTestBench:
    def __init__(self, name: str, uut: Design):
        self._uut = uut
        self.name = name

    def save_to(self, destination: Path):
        test_bench = InProjectTemplate(
            package=module_to_package(self.__module__),
            file_name="lstm_network_tb.tpl.vhd",
            parameters={"name": self.name, "uut_name": self._uut.name},
        )
        destination.create_subpath(f"{self.name}").as_file(".vhd").write(test_bench)



import cocotb
from cocotb.clock import Clock
from cocotb.triggers import RisingEdge, Timer

from elasticai.creator.testing.cocotb_prepare import read_testdata


@cocotb.test()
async def rom_call_test(dut):
    data = read_testdata(dut._name)
    clock_period_ns = 10

    dut.en.value = 0
    dut.clk.value = 0
    dut.addr.value = 0

    cocotb.start_soon(Clock(dut.clk, period=clock_period_ns, units="ns").start())
    await Timer(4 * clock_period_ns, units="ns")
    await RisingEdge(dut.clk)
    dut.en.value = 1

    # --- Getting data
    for addr, data_chck in enumerate(data["data"]):
        dut.addr.value = addr
        await RisingEdge(dut.clk)
        assert dut.ROM.value[addr].signed_integer == data_chck
        await RisingEdge(dut.clk)

    # for addr in range(len(data["data"]), 2**):


@cocotb.test()
async def rom_content_read(dut):
    data = read_testdata(dut._name)
    dut.en.value = 0
    dut.clk.value = 0
    dut.addr.value = 0

    for rom_b, json_b in zip(dut.ROM.value, data["data"]):
        assert rom_b.signed_integer == json_b



from elasticai.creator.ir.base import Attribute, IrData, ReadOnlyField


def _read_only_str() -> ReadOnlyField[str, str]:
    return ReadOnlyField(lambda x: x)


class Node(IrData):
    """
    NOTE: name and type are read only when accessed through their descriptor, i.e.
    >>> n = Node(dict())
    >>> n.name = "x" #  Error! read only
    >>> x = n.name  #  Error! no key `name` in n.data
    >>> n.data['name'] = "x"
    >>> n.name #  Ok!
    ... 'x'
    """

    type: ReadOnlyField[str, str] = _read_only_str()
    __slots__ = ("_name",)

    def __init__(self, name: str, data: dict[str, Attribute]) -> None:
        super().__init__(data)
        self._name = name

    @property
    def name(self) -> str:
        return self._name


class Edge(IrData):
    __slots__ = ("_src", "_dst")

    def __init__(self, src: str, dst: str, data: dict[str, Attribute]) -> None:
        super().__init__(data)
        self._src = src
        self._dst = dst

    @property
    def src(self) -> str:
        return self._src

    @property
    def dst(self) -> str:
        return self._dst


def node(name: str, type: str, attributes: dict[str, Attribute] | None = None) -> Node:
    if attributes is None:
        attributes = dict()
    return Node(name, dict(type=type, **attributes))


def edge(src: str, dst: str, attributes: dict[str, Attribute] | None = None) -> Edge:
    if attributes is None:
        attributes = dict()
    return Edge(src=src, dst=dst, data=attributes)



from collections.abc import Iterable
from typing import cast

import pytest
import torch

from elasticai.creator.nn.fixed_point.precomputed.identity_step_function import (
    IdentityStepFunction,
)
from tests.tensor_test_case import assertTensorEqual


def generate_step_lut(min: float, max: float, steps: int) -> torch.Tensor:
    return torch.linspace(min, max, steps)


@pytest.mark.parametrize(
    "minimum,maximum,steps,inputs,outputs",
    [
        (-3, 3, 2, range(-4, 5), [-3, -3, 3, 3, 3, 3, 3, 3, 3]),
        (-3, 3, 3, range(-4, 5), [-3, -3, 0, 0, 0, 3, 3, 3, 3]),
        (-3, 3, 3, range(5, 10), [3, 3, 3, 3, 3]),
        (-5, -2, 3, range(-6, 0), [-5, -5, -3.5, -2, -2, -2]),
        (2, 5, 3, range(1, 7), [2, 2, 3.5, 5, 5, 5]),
    ],
)
def test_inputs_correctly_mapped_to_step_function_inputs(
    minimum: float,
    maximum: float,
    steps: int,
    inputs: Iterable[float],
    outputs: Iterable[float],
) -> None:
    step_lut = generate_step_lut(minimum, maximum, steps)
    actual_outputs = cast(
        torch.Tensor, IdentityStepFunction.apply(torch.tensor(inputs), step_lut)
    )
    assertTensorEqual(list(outputs), actual_outputs)


@pytest.mark.parametrize("steps", [1, 0])
def test_raises_error_when_steps_less_than_or_equal_one(steps: int) -> None:
    inputs = torch.tensor([1, 2])
    step_lut = generate_step_lut(-1, 1, steps)
    with pytest.raises(ValueError):
        IdentityStepFunction.apply(inputs, step_lut)



from collections.abc import Collection, Iterator, Mapping
from typing import Protocol

from .attribute import Attribute


class _IrData(Protocol):
    _fields: dict[str, type]
    data: dict[str, Attribute]


class _ReadOnlyMappingWithHiddenFields(Mapping[str, Attribute]):
    __slots__ = ("_mapping", "_hidden_fields")

    def __init__(
        self, mapping: Mapping[str, Attribute], hidden_fields: Collection[str]
    ):
        self._mapping = mapping
        self._hidden_fields = hidden_fields

    def __contains__(self, key: object) -> bool:
        return key in self._mapping and key not in self._hidden_fields

    def __len__(self) -> int:
        return len(set(self._mapping.keys()).difference(self._hidden_fields))

    def __iter__(self) -> Iterator[str]:
        for key in self._mapping:
            if key not in self._hidden_fields:
                yield key

    def __getitem__(self, k) -> Attribute:
        if k in self._hidden_fields:
            raise KeyError(f"key not found '{k}'.")
        return self._mapping[k]

    def __or__(self, other: Mapping[str, Attribute]) -> dict[str, Attribute]:
        return dict(self) | dict(other.items())

    def __ror__(self, other: Mapping[str, Attribute]) -> dict[str, Attribute]:
        return dict(other.items()) | dict(self)


class AttributesDescriptor:
    """Read values from the `data` dict, while hiding key, value that show up
    in `_fields` of the owning instance.

    The use for this are cases where mandatory fields are treated in a special
    way and you want write/read access to all other elements from `data`, e.g., while converting an old ir data object into a new one,
    and you just want to get the missing values from `data`.

    NOTE: The returned Mapping is read only
    """

    def __get__(self, instance: _IrData, owner=type[_IrData] | None):
        return _ReadOnlyMappingWithHiddenFields(instance.data, instance._fields)



from abc import ABC, abstractmethod

import torch

from elasticai.creator.vhdl.design.design import Design


class DesignCreatorModule(ABC, torch.nn.Module):
    @abstractmethod
    def create_design(self, name: str) -> Design: ...



import importlib.resources as res
from string import Template

from elasticai.creator.ir2vhdl import (
    Code,
    EntityTemplateDirector,
    Implementation,
    Ir2Vhdl,
    PluginSymbol,
)


def _load_vhd(component: str) -> str:
    vhd_file = res.files(package="elasticai.creator_plugins.skeleton").joinpath(
        f"vhdl/{component}.vhd"
    )
    return vhd_file.read_text()


class _Skeleton(PluginSymbol):
    def __init__(self, template: Template):
        self._template = template

    def __call__(self, arg: Implementation) -> Code:
        def _iter():
            yield self._template.substitute(
                arg.attributes["generic_map"] | dict(entity=arg.name)
            )

        return arg.name, _iter()

    @classmethod
    def load_into(cls, loader: Ir2Vhdl) -> None:
        template = (
            EntityTemplateDirector()
            .set_prototype(_load_vhd("skeleton"))
            .add_generic("DATA_IN_WIDTH")
            .add_generic("DATA_IN_DEPTH")
            .add_generic("DATA_OUT_WIDTH")
            .add_generic("DATA_OUT_DEPTH")
            .build()
        )
        skeleton = cls(template)
        loader.register("skeleton", skeleton)


skeleton = _Skeleton



from elasticai.creator.arithmetic import (
    FxpArithmetic,
    FxpParams,
)
from elasticai.creator.base_modules.hard_sigmoid import HardSigmoid as HardSigmoidBase
from elasticai.creator.nn.design_creator_module import DesignCreatorModule

from .design import HardSigmoid as HardSigmoidDesign


class HardSigmoid(DesignCreatorModule, HardSigmoidBase):
    def __init__(self, total_bits: int, frac_bits: int) -> None:
        """Quantized Activation Function for Sigmoid
        :param total_bits:          Total number of bits
        :param frac_bits:           Fraction of bits
        """
        super().__init__()
        self._params = FxpParams(
            total_bits=total_bits, frac_bits=frac_bits, signed=True
        )
        self._config = FxpArithmetic(self._params)

    def create_design(self, name: str) -> HardSigmoidDesign:
        neg_thr = -3.0
        if self._params.rational_out_of_bounds(neg_thr):
            neg_thr = self._config.minimum_as_rational
        pos_thr = +3.0
        if self._params.rational_out_of_bounds(pos_thr):
            pos_thr = self._config.maximum_as_rational
        y_int = 0.5
        if self._params.rational_out_of_bounds(y_int):
            pos_thr = self._config.maximum_as_rational
        y_max = 1.0
        if self._params.rational_out_of_bounds(y_max):
            y_max = self._config.maximum_as_rational
        slope = 1 / 6
        if self._params.rational_out_of_bounds(slope):
            y_max = self._params.minimum_step_as_rational

        return HardSigmoidDesign(
            name=name,
            total_bits=self._config.total_bits,
            frac_bits=self._config.frac_bits,
            one=self._config.cut_as_integer(y_max),
            zero_threshold=self._config.cut_as_integer(neg_thr),
            one_threshold=self._config.cut_as_integer(pos_thr),
            slope=self._config.cut_as_integer(slope),
            y_intercept=self._config.cut_as_integer(y_int),
        )



import importlib.resources as res

from elasticai.creator.vhdl_template import (
    EntityTemplateDirector,
    Template,
)

from elasticai.creator.ir2vhdl import Code, Implementation, Ir2Vhdl
from elasticai.creator.plugin import PluginSymbol as Loadable


def _load_vhd(component: str) -> str:
    vhd_file = res.files(".").joinpath(f"../vhd/{component}.vhd")
    return vhd_file.read_text()


class _Skeleton(Loadable[Ir2Vhdl]):
    def __init__(self, template: Template):
        self._template = template

    def __call__(self, arg: Implementation) -> Code:
        return "", [""]

    @classmethod
    def load_into(cls, loader: Ir2Vhdl) -> None:
        template = (
            EntityTemplateDirector()
            .set_prototype(_load_vhd("skeleton"))
            .add_generic("DATA_IN_WIDTH")
            .add_generic("DATA_IN_DEPTH")
            .add_generic("DATA_OUT_WIDTH")
            .add_generic("DATA_OUT_DEPTH")
            .build()
        )
        skeleton = cls(template)
        loader.register("skeleton")(skeleton)


skeleton = _Skeleton



from elasticai.creator.arithmetic import (
    FxpArithmetic,
    FxpParams,
)
from elasticai.creator.base_modules.hard_tanh import HardTanh as HardTanhBase
from elasticai.creator.nn.design_creator_module import DesignCreatorModule

from .design import HardTanh as HardTanhDesign


class HardTanh(DesignCreatorModule, HardTanhBase):
    def __init__(
        self,
        total_bits: int,
        frac_bits: int,
        min_val: float = -1.0,
        max_val: float = 1.0,
    ) -> None:
        """Quantized Activation Function for Sigmoid
        :param total_bits:   Total number of bits
        :param frac_bits:    Fraction of bits
        :param min_val:      Floating value of the minimal input/output value (downer limitation)
        :param max_val:      Floating value of the minimal input/output value (upper limitation)
        """
        super().__init__(min_val, max_val)
        self._params = FxpParams(
            total_bits=total_bits, frac_bits=frac_bits, signed=True
        )
        self._config = FxpArithmetic(self._params)

        if self._params.rational_out_of_bounds(min_val):
            self.min_val = self._config.minimum_as_rational
        if self._params.rational_out_of_bounds(max_val):
            self.max_val = self._config.maximum_as_rational

    def create_design(self, name: str) -> HardTanhDesign:
        return HardTanhDesign(
            name=name,
            total_bits=self._config.total_bits,
            frac_bits=self._config.frac_bits,
            min_val=self._config.cut_as_integer(self.min_val),
            max_val=self._config.cut_as_integer(self.max_val),
        )



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information
import os
from importlib.metadata import version as _version
from pathlib import Path

from tomllib import load as _load_toml

project = "elastic-ai.creator"
copyright = "2024, ies-ude (Intelligent Embedded System - University Duisburg-Essen)"
author = "es-ude"
release = _version("elasticai.creator")
version = ".".join(_version("elasticai.creator").split(".")[0:2])
html_title = "Elastic-AI.creator"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "myst_parser",
    "sphinx_copybutton",
    "sphinx_togglebutton",
    "sphinx_book_theme",
    "sphinx.ext.napoleon",
    "sphinx.ext.viewcode",
    "sphinx.ext.todo",
    "sphinx.ext.intersphinx",
    "sphinx.ext.graphviz",
    "sphinx_design",
    "autodoc2",
    "sphinxext.opengraph",
    "sphinxcontrib.plantuml",
    "sphinxcontrib.mermaid",
    "sphinxcontrib.wavedrom",  # TODO: make wavedrom work to render waveforms
]

templates_path = ["_templates"]
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output
# Default theme configuration
html_show_sourcelink = False


# Configure theme based on build type
html_theme = "pydata_sphinx_theme"
html_theme_options = {
    "content_footer_items": ["last-updated"],
    "repository_url": "https://github.com/es-ude/elastic-ai.creator/",
    "path_to_docs": "docs",
    "navigation_depth": 4,
    "icon_links": [
        {
            # Label for this link
            "name": "GitHub",
            # URL where the link will redirect
            "url": "https://github.com/es-ude/elastic-ai.creator/",  # required
            # Icon class (if "type": "fontawesome"), or path to local image (if "type": "local")
            "icon": "fa-brands fa-square-github",
            # The type of image to be used (see below for details)
            "type": "fontawesome",
        }
    ],
}


# only github flavored markdown
myst_gfm_only = False
myst_enable_extensions = [
    "amsmath",
    "attrs_inline",
    "colon_fence",
    "deflist",
    "dollarmath",
    "fieldlist",
    "html_admonition",
    "html_image",
    "linkify",
    "replacements",
    "smartquotes",
    "strikethrough",
    "substitution",
    "tasklist",
]

# allow mermaid usage like on github in markdown
myst_fence_as_directive = ["mermaid", "wavedrom"]

running_in_autobuild = os.getenv("SPHINX_AUTOBUILD", "NO") == "YES"


def find_builtin_plugins():
    plugin_path = Path(__file__).parent / "../elasticai/creator_plugins/"
    source_dir = Path(__file__).parent
    plugins = []
    for plugin in plugin_path.glob("*"):
        if (plugin / "__init__.py").exists():
            plugins.append(plugin)
    result = [
        {
            "path": str(plugin.relative_to(source_dir)),
            "module": f"elasticai.creator_plugins.{plugin.name}",
        }
        for plugin in plugins
    ]
    return result


find_builtin_plugins()

autodoc2_packages = [
    {
        "path": "../elasticai/creator",
        "module": "elasticai.creator",
    },
] + find_builtin_plugins()


def _build_regexes_for_all():
    """These regexes are considered to match for packages/modules that define an __all__ attribute.

    The __all__ attribute will be used to determine the public api of the package/module.
    This function is necessary because autodoc2 will try to build docs even for packages
    that do not define an __all__ attribute and pure namespace packages, however
    that will fail and doc build process aborts. Therefore we explicitly exclude these packages
    from the regexes.
    """
    project_root = Path(__file__).parents[1]

    def find_accessible_module_from_our_architecture():
        modules = []
        with (project_root / "tach.toml").open("r+b") as tach_config:
            architecture_description = _load_toml(tach_config)
        for module in architecture_description["modules"]:
            module_path = module["path"]
            if module_has_an_all_field(module_path):
                modules.append(module_path)
        return modules

    def module_has_an_all_field(module_path):
        file_path = project_root / "/".join(module_path.split("."))
        init_file = file_path / "__init__.py"
        if init_file.exists():
            with init_file.open("r") as f:
                for line in f:
                    if "__all__" in line:
                        return True
        return False

    modules = find_accessible_module_from_our_architecture()
    return modules


autodoc2_module_all_regexes = _build_regexes_for_all()
autodoc2_render_plugin = "myst"
autodoc2_hidden_objects = {"inherited", "private"}

myst_heading_anchors = 3
myst_heading_slug = True



import warnings
from typing import Protocol

from elasticai.creator.file_generation.savable import Path
from elasticai.creator.vhdl.design.design import Design
from elasticai.creator.vhdl.system_integrations.env5_constraints.env5_constraints import (
    ENV5Constraints,
)
from elasticai.creator.vhdl.system_integrations.middleware.middleware import Middleware
from elasticai.creator.vhdl.system_integrations.skeleton.skeleton import (
    EchoSkeletonV2,
    LSTMSkeleton,
)
from elasticai.creator.vhdl.system_integrations.top.env5_reconfig_top import (
    ENV5ReconfigTop,
)

warnings.warn(
    message=(
        "Will be removed. Use LSTMFirmwareENV5 from the"
        " elasticai.creator.vhdl.system_integrations.firmware_env5 package"
        " instead."
    ),
    category=DeprecationWarning,
)


class _DesignAndTestbenchCreator(Protocol):
    def create_design(self, name: str) -> Design: ...

    def create_testbench(self, name: str, design: Design) -> Design: ...


class FirmwareLSTMENv5:
    def __init__(self, model: _DesignAndTestbenchCreator):
        self._network = model.create_design("network")
        self._testbench = model.create_testbench("network_tb", self._network)

    def save_to(self, destination: Path):
        def save_srcs(destination: Path):
            lstm_skeleton = LSTMSkeleton(self._network.name)
            lstm_skeleton.save_to(destination.create_subpath("skeleton"))

            self._network.save_to(destination.create_subpath(self._network.name))

            middleware = Middleware()
            middleware.save_to(destination.create_subpath("middleware"))
            self._testbench.save_to(
                destination.create_subpath("test_benches").create_subpath(
                    self._testbench.name
                )
            )
            env5_reconfig_top = ENV5ReconfigTop()
            env5_reconfig_top.save_to(destination.create_subpath("env5_reconfig_top"))

        def save_constraints(destination: Path):
            env5_config = ENV5Constraints()
            env5_config.save_to(destination.create_subpath("env5_config"))

        srcs = destination.create_subpath("srcs")
        constraints = destination.create_subpath("constraints")
        save_constraints(constraints)
        save_srcs(srcs)


class FirmwareEchoServerSkeletonV2:
    def __init__(self, num_inputs: int, bitwidth: int):
        self._num_inputs = num_inputs
        self._bitwidth = bitwidth
        self.skeleton_id: list[int] = list()

    def save_to(self, destination: Path):
        def save_srcs(destination: Path):
            skeleton = EchoSkeletonV2(self._num_inputs, bitwidth=self._bitwidth)
            self.skeleton_id = skeleton._id
            skeleton.save_to(destination.create_subpath("skeleton"))

            middleware = Middleware()
            middleware.save_to(destination.create_subpath("middleware"))

            env5_reconfig_top = ENV5ReconfigTop()
            env5_reconfig_top.save_to(destination.create_subpath("env5_reconfig_top"))

        def save_constraints(destination: Path):
            env5_config = ENV5Constraints()
            env5_config.save_to(destination.create_subpath("env5_config"))

        srcs = destination.create_subpath("srcs")
        constraints = destination.create_subpath("constraints")
        save_constraints(constraints)
        save_srcs(srcs)



import re


class NameRegistry:
    def __init__(self):
        self._registry = {}

    def _get_name_count(self, name):
        return self._registry.get(name, 0)

    def prepopulate(self, names):
        for name in names:
            match = re.match(r"(.+)_(\d+)$", name)
            suffix = 0
            if match:
                name = match.group(1)
                suffix = int(match.group(2))
            suffix = max(suffix, self._get_name_count(name))
            self._registry[name] = suffix
        return self

    def get_unique_name(self, name):
        if name not in self._registry:
            self._registry[name] = 0
            return name

        new_name = f"{name}_{self._registry[name] + 1}"
        self._registry[name] += 1
        return new_name



from elasticai.creator.arithmetic import FxpConverter, FxpParams
from elasticai.creator.file_generation.savable import Path
from elasticai.creator.file_generation.template import (
    InProjectTemplate,
    module_to_package,
)
from elasticai.creator.vhdl.code_generation.addressable import calculate_address_width


class Rom:
    def __init__(
        self, name: str, data_width: int, values_as_integers: list[int]
    ) -> None:
        self._name = name
        self._data_width = data_width
        self._address_width = calculate_address_width(len(values_as_integers))
        conv = FxpConverter(
            FxpParams(total_bits=self._data_width, frac_bits=0, signed=True)
        )
        self._values = [
            conv.integer_to_binary_string_vhdl(x)
            for x in self._append_zeros_to_fill_addressable_memory(values_as_integers)
        ]

    def save_to(self, destination: Path):
        template = InProjectTemplate(
            file_name="rom.tpl.vhd",
            package=module_to_package(self.__module__),
            parameters=dict(
                rom_value=self._rom_values(),
                rom_addr_bitwidth=str(self._address_width),
                rom_data_bitwidth=str(self._data_width),
                name=self._name,
                resource_option="auto",
            ),
        )
        destination.create_subpath(self._name).as_file(".vhd").write(template)

    def _rom_values(self) -> str:
        return ", ".join(self._values)

    def _append_zeros_to_fill_addressable_memory(self, values: list[int]) -> list[int]:
        missing_number_of_zeros = 2**self._address_width - len(values)
        return values + [0] * missing_number_of_zeros



from collections.abc import Callable
from typing import TypeAlias

import elasticai.creator.plugin as pl
from elasticai.creator.function_utils import FunctionDecorator
from elasticai.creator.ir import LoweringPass as _LoweringPass

from .convolution import DummyLowerable

LoweringPass: TypeAlias = _LoweringPass[DummyLowerable, str]
LoweringFn: TypeAlias = Callable[[DummyLowerable], str]
SymbolFn: TypeAlias = pl.PluginSymbolFn[LoweringPass, [DummyLowerable], str]


def _type_handler(name: str, fn: LoweringFn) -> SymbolFn:
    def load_into(lower: LoweringPass) -> None:
        lower.register(name)(fn)

    return pl.make_plugin_symbol(load_into, fn)


type_handler: FunctionDecorator[LoweringFn, SymbolFn] = FunctionDecorator(_type_handler)


@type_handler
def lowering_pass_conv(x: DummyLowerable) -> str:
    return "_".join(x.more_data)



import pytest

from elasticai.creator.file_generation.on_disk_path import OnDiskPath
from elasticai.creator.nn.fixed_point import Linear
from elasticai.creator.nn.fixed_point.lstm.layer import (
    FixedPointLSTMWithHardActivations,
    LSTMNetwork,
)


@pytest.mark.simulation
def test_lstm_network_simulation(tmp_path):
    hidden_size = 20
    total_bits = 8
    frac_bits = 4
    model = LSTMNetwork(
        [
            FixedPointLSTMWithHardActivations(
                total_bits=total_bits,
                frac_bits=frac_bits,
                input_size=1,
                hidden_size=hidden_size,
                bias=True,
            ),
            Linear(
                total_bits=total_bits,
                frac_bits=frac_bits,
                in_features=hidden_size,
                out_features=1,
                bias=True,
            ),
        ]
    )
    design = model.create_design("lstm_network")
    testbench = model.create_testbench("lstm_network_tb", design)
    testbench.save_to(OnDiskPath(str(tmp_path), parent=""))



import pytest
import torch

from elasticai.creator.arithmetic import FxpParams
from elasticai.creator.nn import fixed_point as nn_creator


@pytest.mark.parametrize(
    "total_bits, frac_bits, num_steps",
    [
        (3, 2, 8),
        (4, 2, 4),
        (4, 3, 4),
        (6, 4, 32),
        (8, 5, 64),
        (10, 6, 16),
        (12, 7, 128),
        (12, 8, 64),
        (12, 10, 512),
    ],
)
def test_prelu_compared_torch(total_bits: int, frac_bits: int, num_steps: int) -> None:
    fxp = FxpParams(total_bits=total_bits, frac_bits=frac_bits, signed=True)
    vrange = (fxp.minimum_as_rational, fxp.maximum_as_rational)
    stimulus = torch.arange(
        start=vrange[0],
        end=vrange[1] + fxp.minimum_step_as_rational,
        step=fxp.minimum_step_as_rational,
    )

    act0 = torch.nn.PReLU()
    out0 = act0(stimulus).detach().numpy()
    act1 = nn_creator.PReLU(
        total_bits=total_bits,
        frac_bits=frac_bits,
        num_steps=num_steps,
        sampling_intervall=vrange,
    )
    out1 = act1(stimulus).detach().numpy()
    metric_mean_abs = float(sum(abs(out1 - out0))) / stimulus.shape[0]
    metric_mean = float(abs(sum(out1 - out0))) / stimulus.shape[0]

    assert metric_mean_abs < 1.5 * fxp.minimum_step_as_rational * (
        out1.max() - out1.min()
    )
    assert metric_mean < 0.7 * fxp.minimum_step_as_rational



from collections.abc import Iterable, Iterator
from pathlib import Path

from elasticai.creator.hw_function_id import HwFunctionIdUpdater as _generalUpdater


def _replace_id_in_vhdl(code: Iterable[str], id: bytes) -> Iterator[str]:
    def split_hex(hex: str):
        for i in range(int(len(hex) / 2)):
            yield hex[i : i + 2]

    def _is_id(line: str):
        return line.strip().startswith("constant SKELETON_ID")

    def _build_skeleton_id_line(id: Iterable[str]) -> str:
        id_str = ", ".join(f'x"{b.upper()}"' for b in id)

        return f"  constant SKELETON_ID : skeleton_id_t := ({id_str});"

    digest_as_hex_array = tuple(split_hex(id.hex()))

    for line in code:
        if _is_id(line):
            yield _build_skeleton_id_line(digest_as_hex_array)
            yield "\n"
        else:
            yield line


class HwFunctionIdUpdater:
    def __init__(
        self,
        build_dir: Path,
        target_file: str | Path,
    ):
        self._updater = _generalUpdater(build_dir, target_file, _replace_id_in_vhdl)

    def compute_id(self) -> None:
        self._updater.compute_id()

    def write_id(self) -> None:
        self._updater.write_id()

    @property
    def id(self) -> bytes:
        return self._updater.id



from types import SimpleNamespace

import pytest

from elasticai.creator.function_utils import KeyedFunctionDispatcher
from elasticai.creator.ir import LoweringPass

"""
Tests:

For later in LoweringPass:
    - convert function into one that returns iterable in case it is not already
"""


def test_calls_a_registered_function():
    r = KeyedFunctionDispatcher(dispatch_key_fn=lambda x: x.type)
    c = SimpleNamespace(name="c0", type="convolution")

    @r.register
    def convolution(x) -> str:
        return f"conv{x.name}"

    assert "convc0" == r.call(c)


def test_calling_a_function_with_custom_name():
    r = KeyedFunctionDispatcher(dispatch_key_fn=lambda x: x.type)
    c = SimpleNamespace(name="c0", type="convolution")

    @r.register("convolution")
    def fn(x) -> str:
        return f"conv{x.name}"

    assert "convc0" == r.call(c)


def test_register_lowering_fn_with_fn_name() -> None:
    c = SimpleNamespace(name="c0", type="convolution")
    p: LoweringPass[SimpleNamespace, str] = LoweringPass()

    @p.register
    def convolution(x: SimpleNamespace) -> str:
        return f"conv{x.name}"

    assert ("convc0",) == tuple(p((c,)))


def test_can_directly_call_fn_as_defined() -> None:
    c = SimpleNamespace(name="c0", type="convolution")
    p: LoweringPass[SimpleNamespace, str] = LoweringPass()

    @p.register
    def convolution(x: SimpleNamespace) -> str:
        return f"conv{x.name}"

    assert "convc0" == convolution(c)


def test_register_lowering_fn_with_custom_name() -> None:
    c = SimpleNamespace(name="c0", type="convolution")
    p: LoweringPass[SimpleNamespace, str] = LoweringPass()

    @p.register("convolution")
    def f(x: SimpleNamespace) -> str:
        return f"conv{x.name}"

    assert ("convc0",) == tuple(p((c,)))


def test_lowering_iterates_over_input() -> None:
    c = SimpleNamespace(name="c0", type="convolution")
    d = SimpleNamespace(name="c1", type="convolution")
    p: LoweringPass[SimpleNamespace, str] = LoweringPass()

    @p.register
    def convolution(x: SimpleNamespace) -> str:
        return f"conv{x.name}"

    assert {"convc0", "convc1"} == set(p((c, d)))


def test_can_register_and_call_iter_fn() -> None:
    c = SimpleNamespace(name="c0", type="convolution")
    p: LoweringPass[SimpleNamespace, str] = LoweringPass()

    @p.register_iterable
    def convolution(x: SimpleNamespace) -> tuple[str, ...]:
        return f"conv{x.name}", "more content"

    assert {"convc0", "more content"} == set(p((c,)))


def test_registering_a_name_as_iterable_and_non_iterable_yields_error() -> None:
    p: LoweringPass = LoweringPass()

    @p.register
    def a(x):
        return x

    with pytest.raises(
        ValueError, match="function for a already defined in lowering pass"
    ):

        @p.register_iterable("a")
        def b(x):
            return (x,)



import copy
from collections.abc import Sequence
from typing import Protocol, TypeVar, cast

from elasticai.creator import graph as g
from elasticai.creator import ir
from elasticai.creator.graph import Graph
from elasticai.creator.torch2ir import Implementation


def build_sequential_pattern(
    sequence: Sequence[ir.Node],
) -> ir.Implementation[ir.Node, ir.Edge]:
    pattern = ir.Implementation(graph=g.BaseGraph())
    previous = None
    for node in sequence:
        pattern.add_node(node)
        if previous is not None:
            pattern.add_edge(src=previous, dst=node.name, data={})
        previous = node.name

    return pattern


PNode = TypeVar("PNode", bound=ir.Node)
GNode = TypeVar("GNode", bound=ir.Node)

PNodeCon = TypeVar("PNodeCon", bound=ir.Node, contravariant=True)
GNodeCon = TypeVar("GNodeCon", bound=ir.Node, contravariant=True)


class NodeConstraint(Protocol[PNodeCon, GNodeCon]):
    def __call__(self, *, pattern_node: PNodeCon, graph_node: GNodeCon) -> bool: ...


class _Matcher:
    def __init__(
        self,
        pattern: ir.Implementation[ir.Node, ir.Edge],
        graph: ir.Implementation[ir.Node, ir.Edge],
        node_constraint: NodeConstraint[ir.Node, ir.Node],
    ):
        self.pattern = pattern
        self.graph = graph
        self._node_constraint = node_constraint

    def set_graph(self, graph: ir.Implementation[ir.Node, ir.Edge]) -> None:
        self.graph = graph

    def node_constraint(self, pattern_node: str, graph_node: str) -> bool:
        return self._node_constraint(
            pattern_node=self.pattern.nodes[pattern_node],
            graph_node=self.graph.nodes[graph_node],
        )

    def __call__(self, pattern: Graph[str], graph: Graph[str]) -> dict[str, str]:
        return g.find_subgraph(
            pattern=pattern, graph=graph, node_constraint=self.node_constraint
        )


class SequenceReorderer:
    def __init__(
        self,
        old_order: Sequence[ir.Node],
        new_order: Sequence[ir.Node],
        node_constraint: NodeConstraint[ir.Node, ir.Node],
    ):
        """Use this to reorder the first occurence of a pattern as specified by the replacement."""
        pattern = old_order
        replacement = new_order
        self.pattern = build_sequential_pattern(pattern)
        self.replacement = build_sequential_pattern(replacement)
        self.lhs = {"start": pattern[0].name, "end": pattern[-1].name}
        self.rhs = {"start": replacement[0].name, "end": replacement[-1].name}
        self.matcher = _Matcher(
            pattern=self.pattern,
            graph=Implementation(data={}, graph=g.BaseGraph()),
            node_constraint=node_constraint,
        )
        self._has_changed = False

    def _get_nodes_that_are_in_pattern_and_replacement(self) -> set[str]:
        return set(self.pattern.nodes) & set(self.replacement.nodes)

    def has_changed(self) -> bool:
        """Tells whether the graph has been changed by the last call to reorder."""
        return self._has_changed

    def reorder(self, impl: ir.Implementation) -> ir.Implementation:
        """Reorder the first occurence of the pattern as specified by the replacement.

        The resulting graph and its data are deep copies of the input graph and its data.
        The new node names are unique but chosen based on the replacement nodes and
        not based on the original node names.
        Use `has_changed` to check whether the graph has been changed by the last call to `reorder`
        if you want to keep reordering until you reach a fixed point.

        :::{note}
        Will create a deep copy even if no changes were made.
        As such it is always safe to `del original_graph` after calling `reorder`.
        :::
        """
        self.matcher.set_graph(impl)
        match = self.matcher(self.pattern.graph, impl.graph)
        new_graph, new_names = g.rewrite(
            original=impl.graph,
            replacement=self.replacement.graph,
            match=match,
            lhs=self.lhs,
            rhs=self.rhs,
        )
        self._has_changed = len(match) > 0

        new_data = copy.deepcopy(impl.data)

        def copy_original_data_to_replaced_subgraph(
            pattern_node: str, replacement_node: str
        ) -> None:
            new_name = new_names[replacement_node]

            cast(dict[str, ir.Attribute], new_data["nodes"])[new_name] = cast(
                dict,
                copy.deepcopy(
                    cast(dict[str, ir.Attribute], new_data["nodes"])[
                        cast(str, match[pattern_node])
                    ]
                ),
            )

        for node in self._get_nodes_that_are_in_pattern_and_replacement():
            copy_original_data_to_replaced_subgraph(node, node)
        return ir.Implementation(graph=new_graph, data=new_data)



from collections.abc import Iterable
from functools import partial
from typing import Any, cast

import numpy as np

from elasticai.creator.arithmetic import (
    FxpArithmetic,
    FxpParams,
)

from ._common_imports import (
    Design,
    InProjectTemplate,
    Path,
    Port,
    Rom,
    Signal,
    calculate_address_width,
    module_to_package,
    std_signals,
)


class FPLSTMCell(Design):
    def __init__(
        self,
        *,
        name: str,
        hardtanh: Design,
        hardsigmoid: Design,
        total_bits: int,
        frac_bits: int,
        w_ih: list[list[list[int]]],
        w_hh: list[list[list[int]]],
        b_ih: list[list[int]],
        b_hh: list[list[int]],
    ) -> None:
        super().__init__(name=name)
        work_library_name: str = "work"

        self.input_size = len(w_ih[0])
        self.hidden_size = len(w_ih) // 4
        self.weights_ih = w_ih
        self.weights_hh = w_hh
        self.biases_ih = b_ih
        self.biases_hh = b_hh
        self._params = FxpParams(
            total_bits=total_bits, frac_bits=frac_bits, signed=True
        )
        self._config = FxpArithmetic(self._params)
        self._htanh = hardtanh
        self._hsigmoid = hardsigmoid
        self._rom_base_names = ("wi", "wf", "wg", "wo", "bi", "bf", "bg", "bo")
        self._ram_base_name = f"dual_port_2_clock_ram_{self.name}"
        self._template = InProjectTemplate(
            package=module_to_package(self.__module__),
            file_name=f"{self.name}.tpl.vhd",
            parameters=dict(
                name=self.name,
                library=work_library_name,
                tanh_name=self._htanh.name,
                sigmoid_name=self._hsigmoid.name,
                data_width=str(total_bits),
                frac_width=str(frac_bits),
                input_size=str(self.input_size),
                hidden_size=str(self.hidden_size),
                x_h_addr_width=str(
                    calculate_address_width(self.input_size + self.hidden_size)
                ),
                hidden_addr_width=str(calculate_address_width(self.hidden_size)),
                w_addr_width=str(
                    calculate_address_width(
                        (self.input_size + self.hidden_size) * self.hidden_size
                    )
                ),
            ),
        )

    @property
    def total_bits(self) -> int:
        return int(cast(str, self._template.parameters["data_width"]))

    @property
    def frac_bits(self) -> int:
        return int(cast(str, self._template.parameters["frac_width"]))

    @property
    def _hidden_addr_width(self) -> int:
        return int(cast(str, self._template.parameters["hidden_addr_width"]))

    @property
    def _weight_address_width(self) -> int:
        return int(cast(str, self._template.parameters["w_addr_width"]))

    @property
    def port(self) -> Port:
        ctrl_signal = partial(Signal, width=0)
        return Port(
            incoming=[
                std_signals.clock(),
                # ctrl_signal("clk_hadamard"),
                ctrl_signal("reset"),
                std_signals.enable(),
                ctrl_signal("zero_state"),
                Signal("x_data", width=self.total_bits),
                ctrl_signal("h_out_en"),
                Signal("h_out_addr", width=self._hidden_addr_width),
            ],
            outgoing=[
                std_signals.done(),
                Signal("h_out_data", self.total_bits),
            ],
        )

    def get_file_load_order(self) -> list[str]:
        return [f"{file}.vhd" for file in self._get_qualified_rom_names()] + [
            f"{self._ram_base_name}.vhd",
            f"{self._htanh.name}.vhd",
            f"{self._hsigmoid.name}.vhd",
        ]

    def save_to(self, destination: Path) -> None:
        destination = destination.create_subpath(self.name)
        weights, biases = self._build_weights()

        self._save_roms(
            destination=destination,
            parameters=[*weights, *biases],
        )
        self._save_dual_port_double_clock_ram(destination)
        self._save_hardtanh(destination)
        self._save_sigmoid(destination)

        destination.create_subpath("lstm_cell").as_file(".vhd").write(self._template)

    def _build_weights(self) -> tuple[list[list], list[list]]:
        weights = np.concatenate(
            (np.array(self.weights_ih), np.array(self.weights_hh)), axis=1
        )
        w_i, w_f, w_g, w_o = weights.reshape(4, -1).tolist()

        bias = np.add(self.biases_ih, self.biases_hh)
        b_i, b_f, b_g, b_o = bias.reshape(4, -1).tolist()

        return [w_i, w_f, w_g, w_o], [b_i, b_f, b_g, b_o]

    def _get_qualified_rom_names(self) -> list[str]:
        suffix = f"_rom_{self.name}"
        return [name + suffix for name in self._rom_base_names]

    def _save_roms(self, destination: Path, parameters: Iterable[Any]) -> None:
        for name, values in zip(self._get_qualified_rom_names(), parameters):
            rom = Rom(
                name=name,
                data_width=self.total_bits,
                values_as_integers=values,
            )
            rom.save_to(destination)

    def _save_hardtanh(self, destination: Path) -> None:
        self._htanh.save_to(destination)

    def _save_sigmoid(self, destination: Path) -> None:
        self._hsigmoid.save_to(destination)

    def _save_dual_port_double_clock_ram(self, destination: Path) -> None:
        template = InProjectTemplate(
            file_name="dual_port_2_clock_ram.tpl.vhd",
            package=module_to_package(self.__module__),
            parameters=dict(name=self.name),
        )
        destination.create_subpath(self._ram_base_name).as_file(".vhd").write(template)



from elasticai.creator_plugins.combinatorial.wiring import connect_data_signals


def test_connecting_explicit_data_signals():
    edge = ("src", "dst", ((1, 1),))
    code = tuple(connect_data_signals((edge,)))
    assert code == ("d_in_dst(1) <= d_out_src(1);",)


def test_connecting_data_signals_by_range():
    edge = ("src", "dst", ("range(0, 12)", "range(1, 13)"))
    code = tuple(connect_data_signals((edge,)))
    assert code == ("d_in_dst(12 downto 1) <= d_out_src(11 downto 0);",)



from typing import cast

from elasticai.creator.file_generation.in_memory_path import InMemoryFile, InMemoryPath
from elasticai.creator.nn.identity.layer import BufferedIdentity, BufferlessIdentity


def test_buffered_identity_generates_correct_vhdl_code() -> None:
    expected = """library ieee;
use ieee.std_logic_1164.all;


entity identity is
    port (
        enable : in std_logic;
        clock  : in std_logic;
        x_address : out std_logic_vector(3-1 downto 0);
        y_address : in std_logic_vector(3-1 downto 0);
        x   : in std_logic_vector(16-1 downto 0);
        y  : out std_logic_vector(16-1 downto 0);
        done   : out std_logic
    );
end identity;

architecture rtl of identity is
begin
    y <= x;
    done <= enable;
    x_address <= y_address;
end rtl;
""".splitlines()
    identity = BufferedIdentity(num_input_features=6, total_bits=16)
    build_path = InMemoryPath("build", parent=None)
    design = identity.create_design("identity")
    design.save_to(build_path)
    actual = cast(InMemoryFile, build_path["identity"]).text
    assert actual == expected


def test_bufferless_identity_generates_correct_vhdl_code() -> None:
    expected = """library ieee;
use ieee.std_logic_1164.all;


entity identity is
    port (
        enable : in std_logic;
        clock  : in std_logic;
        x   : in std_logic_vector(8-1 downto 0);
        y  : out std_logic_vector(8-1 downto 0);
    );
end identity;

architecture rtl of identity is
begin
    y <= x;
end rtl;
""".splitlines()
    identity = BufferlessIdentity(total_bits=8)
    build_path = InMemoryPath("build", parent=None)
    design = identity.create_design("identity")
    design.save_to(build_path)
    actual = cast(InMemoryFile, build_path["identity"]).text
    assert actual == expected



from elasticai.creator.ir2verilog import TemplateDirector


class TestVerilogTemplate:
    def test_can_set_localparam(self):
        template = (
            TemplateDirector()
            .set_prototype("localparam DATA_WIDTH = 4'd8;")
            .localparam("DATA_WIDTH")
            .build()
        )

        assert (
            template.substitute({"DATA_WIDTH": "'h16"})
            == "localparam DATA_WIDTH = 'h16;"
        )

    def test_can_set_parameter_before_comma(self):
        template = (
            TemplateDirector()
            .set_prototype("parameter DATA_WIDTH = 4'd8,")
            .parameter("DATA_WIDTH")
            .build()
        )

        assert (
            template.substitute({"DATA_WIDTH": "'h16"})
            == "parameter DATA_WIDTH = 'h16,"
        )

    def test_can_set_parameter_before_new_line(self):
        template = (
            TemplateDirector()
            .set_prototype("""parameter DATA_WIDTH = 4'd8""")
            .parameter("DATA_WIDTH")
            .build()
        )
        assert (
            template.substitute({"DATA_WIDTH": "'h16"}) == "parameter DATA_WIDTH = 'h16"
        )

    def test_can_define(self):
        template = (
            TemplateDirector()
            .set_prototype("//`define DATA_WIDTH")
            .define_scoped_switch("DATA_WIDTH", True)
            .build()
        )
        assert template.substitute({}) == "`define DATA_WIDTH"

    def test_can_undefine(self):
        template = (
            TemplateDirector()
            .set_prototype("`define DATA_WIDTH")
            .define_scoped_switch("DATA_WIDTH", False)
            .build()
        )
        assert template.substitute({}) == "//`define DATA_WIDTH"

    def test_can_leave_defined(self):
        template = (
            TemplateDirector()
            .set_prototype("`define DATA_WIDTH")
            .define_scoped_switch("DATA_WIDTH", True)
            .build()
        )
        assert template.substitute({}) == "`define DATA_WIDTH"

    def test_can_leave_undefined(self):
        template = (
            TemplateDirector()
            .set_prototype("//`define DATA_WIDTH")
            .define_scoped_switch("DATA_WIDTH", False)
            .build()
        )
        assert template.substitute({}) == "//`define DATA_WIDTH"

    def test_can_define_data_width(self):
        template = (
            TemplateDirector()
            .set_prototype("// `define DATA_WIDTH 8")
            .define_scoped_switch("DATA_WIDTH", True)
            .build()
        )
        assert template.substitute({}) == "`define DATA_WIDTH 8"

    def test_can_undefine_data_width(self):
        template = (
            TemplateDirector()
            .set_prototype("`define DATA_WIDTH 8")
            .define_scoped_switch("DATA_WIDTH", False)
            .build()
        )
        assert template.substitute({}) == """//`define DATA_WIDTH 8"""

    def test_can_leave_data_width_defined(self):
        template = (
            TemplateDirector()
            .set_prototype("`define DATA_WIDTH 8")
            .define_scoped_switch("DATA_WIDTH", True)
            .build()
        )
        assert template.substitute({}) == "`define DATA_WIDTH 8"

    def test_can_leave_data_width_undefined(self):
        template = (
            TemplateDirector()
            .set_prototype("//`define DATA_WIDTH 8")
            .define_scoped_switch("DATA_WIDTH", False)
            .build()
        )
        assert template.substitute({}) == "//`define DATA_WIDTH 8"

    def test_can_replace_module_name(self):
        template = (
            TemplateDirector()
            .set_prototype("module FILTER_FIR_HALF#(")
            .add_module_name()
            .build()
        )
        assert (
            template.substitute({"module_name": "FILTER_FIR_MY_NAME_HALF"})
            == "module FILTER_FIR_MY_NAME_HALF#("
        )

    def test_automatically_use_module_prefix_for_defines(self):
        template = (
            TemplateDirector()
            .set_prototype("`define DATA_WIDTH 8")
            .define_scoped_switch("DATA_WIDTH", True)
            .build()
        )
        assert (
            template.substitute({"module_name": "FILTER_FIR_HALF"})
            == "`define FILTER_FIR_HALF_DATA_WIDTH 8"
        )

    def test_later_occurences_of_defined_names_are_scoped(self):
        template = (
            TemplateDirector()
            .set_prototype("`define DATA_WIDTH 8\nDATA_WIDTH + 10")
            .define_scoped_switch("DATA_WIDTH", True)
            .build()
        )
        assert (
            template.substitute({"module_name": "FILTER_FIR_HALF"})
            == "`define FILTER_FIR_HALF_DATA_WIDTH 8\nFILTER_FIR_HALF_DATA_WIDTH + 10"
        )

    def test_can_replace_arrays(self):
        template = (
            TemplateDirector()
            .set_prototype("localparam DATA = {'d2, 'd3};")
            .localparam("DATA")
            .build()
        )
        assert template.substitute({"DATA": "'d5"}) == "localparam DATA = 'd5;"

    def test_replace_module_of_instance_with_parameter(self):
        template = (
            TemplateDirector()
            .set_prototype("""FILT_BIQUAD#(BITWIDTH_DATA, LENGTH) DUT(
    .CLK(clk_sys),
	  .nRST(nrst),
	  .EN(en_dut),
	  .START_FLAG(clk_adc),
	  .DATA_IN(filter_in),
	  .DATA_OUT(dout),
	  .DATA_VALID(filter_rdy)
);""")
            .replace_module_of_instance("FILT_BIQUAD", "MyModule")
            .build()
        )
        assert (
            template.substitute({})
            == """FILT_BIQUAD#(BITWIDTH_DATA, LENGTH) MyModule(
    .CLK(clk_sys),
	  .nRST(nrst),
	  .EN(en_dut),
	  .START_FLAG(clk_adc),
	  .DATA_IN(filter_in),
	  .DATA_OUT(dout),
	  .DATA_VALID(filter_rdy)
);"""
        )

    def test_replace_module_of_instance_without_parameter(self):
        template = (
            TemplateDirector()
            .set_prototype("""FILT_BIQUAD DUT(
    .CLK(clk_sys),
    .nRST(nrst),
    .EN(en_dut),
    .START_FLAG(clk_adc),
    .DATA_IN(filter_in),
    .DATA_OUT(dout),
    .DATA_VALID(filter_rdy)
);""")
            .replace_module_of_instance("FILT_BIQUAD", "MyModule")
            .build()
        )
        assert (
            template.substitute({})
            == """FILT_BIQUAD MyModule(
    .CLK(clk_sys),
    .nRST(nrst),
    .EN(en_dut),
    .START_FLAG(clk_adc),
    .DATA_IN(filter_in),
    .DATA_OUT(dout),
    .DATA_VALID(filter_rdy)
);"""
        )

    def test_replace_module_of_instance_with_space(self):
        template = (
            TemplateDirector()
            .set_prototype("""   FILT_BIQUAD #(4) DUT (
    .CLK(clk_sys),
	  .nRST(nrst),
	  .EN(en_dut),
	  .START_FLAG(clk_adc),
	  .DATA_IN(filter_in),
	  .DATA_OUT(dout),
	  .DATA_VALID(filter_rdy)
);""")
            .replace_module_of_instance("FILT_BIQUAD", "MyModule")
            .build()
        )
        assert (
            template.substitute({})
            == """FILT_BIQUAD #(4) MyModule (
    .CLK(clk_sys),
	  .nRST(nrst),
	  .EN(en_dut),
	  .START_FLAG(clk_adc),
	  .DATA_IN(filter_in),
	  .DATA_OUT(dout),
	  .DATA_VALID(filter_rdy)
);"""
        )

    def test_replace_instance_name_of_module(self):
        template = (
            TemplateDirector()
            .set_prototype("""   FILT_BIQUAD#(4) DUT (
    .CLK(clk_sys),
	  .nRST(nrst),
	  .EN(en_dut),
	  .START_FLAG(clk_adc),
	  .DATA_IN(filter_in),
	  .DATA_OUT(dout),
	  .DATA_VALID(filter_rdy)
);""")
            .replace_instance_name("FILT_BIQUAD", "MyModule")
            .build()
        )
        assert (
            template.substitute({})
            == """   MyModule#(4) DUT (
    .CLK(clk_sys),
	  .nRST(nrst),
	  .EN(en_dut),
	  .START_FLAG(clk_adc),
	  .DATA_IN(filter_in),
	  .DATA_OUT(dout),
	  .DATA_VALID(filter_rdy)
);"""
        )

    def test_replace_instance_name_of_module_with_space(self):
        template = (
            TemplateDirector()
            .set_prototype("""  FILT_BIQUAD #(4) DUT (
    .CLK(clk_sys),
	  .nRST(nrst),
	  .EN(en_dut),
	  .START_FLAG(clk_adc),
	  .DATA_IN(filter_in),
	  .DATA_OUT(dout),
	  .DATA_VALID(filter_rdy)
);""")
            .replace_instance_name("FILT_BIQUAD", "MyModule")
            .build()
        )
        assert (
            template.substitute({})
            == """  MyModule #(4) DUT (
    .CLK(clk_sys),
	  .nRST(nrst),
	  .EN(en_dut),
	  .START_FLAG(clk_adc),
	  .DATA_IN(filter_in),
	  .DATA_OUT(dout),
	  .DATA_VALID(filter_rdy)
);"""
        )



from .autograd import get_autograd_func
from .module_quantization import (
    QuantizeForwHTE,
    QuantizeForwHTEBackwHTE,
    QuantizeForwHTEBackwStochastic,
    QuantizeForwStochastic,
    QuantizeForwStochasticBackwStochastic,
    get_fxp_forw_quantization_module,
    get_fxp_forwbackw_quantization_module,
)
from .param_quantization import (
    QuantizeParamSTEToFixedPointHTE,
    QuantizeParamSTEToFixedPointStochastic,
    QuantizeParamToFixedPointHTE,
    QuantizeParamToFixedPointStochastic,
)
from .quantize_to_fixed_point import quantize_to_fxp_hte, quantize_to_fxp_stochastic
from .two_complement_fixed_point_config import FixedPointConfigV2

__all__ = [
    "quantize_to_fxp_hte",
    "quantize_to_fxp_stochastic",
    "get_autograd_func",
    "get_fxp_forwbackw_quantization_module",
    "get_fxp_forw_quantization_module",
    "QuantizeForwHTE",
    "QuantizeForwHTEBackwHTE",
    "QuantizeForwHTEBackwStochastic",
    "QuantizeForwStochastic",
    "QuantizeForwStochasticBackwStochastic",
    "QuantizeParamSTEToFixedPointHTE",
    "QuantizeParamSTEToFixedPointStochastic",
    "QuantizeParamToFixedPointHTE",
    "QuantizeParamToFixedPointStochastic",
    "FixedPointConfigV2",
    "QuantizeTensorToFixedPointHTE",
    "QuantizeTensorToFixedPointStochastic",
]



from collections.abc import Callable
from typing import Generic, TypeVar, cast

from typing_extensions import TypeIs

from ._has_data import HasData
from .attribute import Attribute

StoredT = TypeVar("StoredT", bound=Attribute)
VisibleT = TypeVar("VisibleT")


class RequiredField(Generic[StoredT, VisibleT]):
    """
    A __descriptor__ that designates a mandatory field of an abstract ir data class.
    The descriptor accesses the `data` dictionary of the owning abstract ir data object
    to read and write values. You can use the `set_convert` and `get_convert` functions
    to transform values during read/write accesses.
    That allows to keep dictionary of primitive (serializable) data types in memory,
    while still providing abstract ways to manipulate that data in complex ways.
    This is typically required when working with Nodes and Graphs to create new
    intermediate representations and transform one graph into another.

    NOTE: In most cases you want to ensure that the conditions
      - `x == get_convert(set_convert(x))`
      - `x == set_convert(get_convert(x))`
    hold for all applicable `x`.

    E.g.
    ```python
    >>> from elasticai.creator.ir import IrData
    >>> class A(IrData):
    ...    number: RequiredField[str, int] = RequiredField(set_transform=str, get_transform=int)
    >>> a = A({'number': "12"})
    >>> a.number = a.number + 3
    >>> a.data
    {'number': "15"}
    ```
    """

    __slots__ = ("set_convert", "get_convert", "name")

    def __init__(
        self,
        set_convert: Callable[[VisibleT], StoredT],
        get_convert: Callable[[StoredT], VisibleT],
    ):
        self.set_convert = set_convert
        self.get_convert = get_convert

    def __set_name__(self, owner: type[HasData], name: str) -> None:
        """
        IMPORTANT: do not remove owner even though it's not used
        see https://docs.python.org/3/reference/datamodel.html#descriptors for more information
        """
        self.name = name

    def __get__(self, instance: HasData, owner=None) -> VisibleT:
        """
        IMPORTANT: do not remove owner even though it's not used
        see https://docs.python.org/3/reference/datamodel.html#descriptors for more information
        """
        return self.get_convert(cast(StoredT, instance.data[self.name]))

    def __set__(self, instance: HasData, value: VisibleT) -> None:
        instance.data[self.name] = self.set_convert(value)


class SimpleRequiredField(Generic[StoredT]):
    slots = ("name",)

    def __set_name__(self, owner: type[HasData], name: str) -> None:
        self.name = name

    def __get__(self, instance: HasData, owner=None) -> StoredT:
        return cast(StoredT, instance.data[self.name])

    def __set__(self, instance: HasData, value: StoredT) -> None:
        instance.data[self.name] = value


class ReadOnlyField(Generic[StoredT, VisibleT]):
    slots = ("get_convert",)

    def __init__(self, get_convert: Callable[[StoredT], VisibleT]) -> None:
        self.get_convert = get_convert

    def __set_name__(self, owner: type[HasData], name: str) -> None:
        self.name = name

    def __get__(
        self, instance: HasData, owner: type[HasData] | None = None
    ) -> VisibleT:
        return self.get_convert(cast(StoredT, instance.data[self.name]))


_HasDataT = TypeVar("_HasDataT", bound=HasData)


class ReadOnlyMethodField(Generic[_HasDataT, StoredT, VisibleT]):
    __slots__ = ("get_convert", "name")

    def __init__(self, get_convert: Callable[[_HasDataT, StoredT], VisibleT]) -> None:
        self.get_convert = get_convert
        self.name: str = "<not set>"

    def __set_name__(self, owner: type, name: str) -> None:
        if self.name == "<not set>":
            self.name = name

    def __get__(self, instance: _HasDataT, owner=None) -> VisibleT:
        return self.get_convert(instance, cast(StoredT, instance.data[self.name]))


class StaticMethodField(Generic[StoredT, VisibleT]):
    __slots__ = ("set_convert", "get_convert", "name")

    def __init__(self, get_convert: Callable[[StoredT], VisibleT]) -> None:
        self.get_convert = get_convert
        self.name: str = "<not set>"

        def set_convert(value: VisibleT) -> StoredT:
            raise NotImplementedError

        self.set_convert: Callable[[VisibleT], StoredT] = set_convert

    def __set_name__(self, owner: type, name: str) -> None:
        if self.name == "<not set>":
            self.name = name

    def __get__(self, instance: _HasDataT, owner=None) -> VisibleT:
        return self.get_convert(cast(StoredT, instance.data[self.name]))

    def __set__(self, instance: _HasDataT, value: VisibleT) -> None:
        instance.data[self.name] = self.set_convert(value)

    def setter(self, fn: Callable[[VisibleT], StoredT]) -> None:
        self.set_convert = fn


def read_only_field(
    fn: Callable[[_HasDataT, StoredT], VisibleT],
) -> ReadOnlyMethodField[_HasDataT, StoredT, VisibleT]:
    """Decorate a method as getter for a read only field.

    This works similar to the `property` decorator, but will
    automatically pass the content of the `.data` dictionary
    to the decorated method, that matches its name, e.g.,
    decorating a method `name` will call it with `self.data['name']`.
    The method will also be bound as an instance method, i.e.,
    the owning instance will be passed as the first argument.

    Additionally, the name of the decorated method will be registered
    as a required field for all other purposes.

    :::{admonition} Example
    ```python
    from elasticai.creator.ir import IrData, required_field

    class MyData(IrData):
        def __init__(self, data: dict[str, Attribute]):
            self.data = data
            self.length_scaling = 2

        @read_only_field
        def length(self, value: str) -> str:
            return value
    ```
    :::
    For more information see [`required_field`](#elasticai.creator.ir.required_field).
    """
    return ReadOnlyMethodField(fn)


def static_required_field(
    fn: Callable[[StoredT], VisibleT],
) -> StaticMethodField[StoredT, VisibleT]:
    """Decorate a static method as getter for a required field.

    Opposed to `read_only_field` this decorator will not pass the owning instance
    to the decorated method. This is to avoid hard to catch inconsistencies
    if the owning instances state changes between read and write operations.

    The main purpose of this decorator is to providea a more readable alternative
    to [`RequiredField`](#elasticai.creator.ir.required_field.RequiredField).

    :::{note}
    Type checkers will not be able to pickup that the decorated methods are static by means of our
    custom descriptor. That means you will probably have to decorate the method with `@staticmethod`
    before applying `@static_required_field`.
    :::

    :::{admonition} Example
    ```python
    from elasticai.creator.ir import IrData, static_required_field

    class MyData(IrData):
        @static_required_field
        @staticmethod
        def length(value: int) -> float:
            return value * 2.0

        @length.setter
        @staticmethod
        def _(value: float) -> int:
            return int(value / 2.0)
    :::

    See also [`read_only_field`](#elasticai.creator.ir.required_field.read_only_field).
    """
    return StaticMethodField(fn)


_required_fields = [
    RequiredField,
    ReadOnlyField,
    SimpleRequiredField,
    StaticMethodField,
    ReadOnlyMethodField,
]


def is_required_field(
    o: object,
) -> TypeIs[RequiredField | ReadOnlyField | SimpleRequiredField]:
    def _isinstance(t):
        return isinstance(o, t)

    return any(map(_isinstance, _required_fields))


def is_required_field_type(
    cls: type,
) -> TypeIs[type[RequiredField] | type[ReadOnlyField] | type[RequiredField]]:
    def _issubclass(t):
        return isinstance(cls, t)

    return any(map(_issubclass, _required_fields))


def register_required_field_type(
    cls: type,
) -> None:
    if cls not in _required_fields:
        _required_fields.append(cls)



from elasticai.creator.ir2vhdl import Shape, VhdlNode
from elasticai.creator_plugins.combinatorial.vhdl_nodes.striding_shift_register import (
    striding_shift_register,
)

from .shift_register_test import extract_code_section


def test_shift_register_converts_depth_and_width_to_correct_generics():
    conv0_channels = 2
    conv1_kernel_size = 3
    conv0_out_shape = Shape(conv0_channels, 1)
    conv1_in_shape = Shape(conv0_channels, conv1_kernel_size)
    stride = 2
    n = VhdlNode(
        name="sr0",
        data=dict(
            type="shift_register",
            implementation="",
            input_shape=conv0_out_shape.to_tuple(),
            output_shape=conv1_in_shape.to_tuple(),
            stride=stride,
        ),
    )
    sr = striding_shift_register(n)
    entity = tuple(extract_code_section(sr.instantiate(), start="generic", end=")"))
    entity = set(line.strip(",") for line in entity)
    expected = {
        f"DATA_WIDTH => {conv0_channels}",
        f"NUM_POINTS => {conv1_in_shape.width}",
        f"STRIDE => {stride}",
    }
    assert entity == expected



from collections.abc import Callable, Iterable, Iterator, Sequence
from dataclasses import dataclass
from importlib import resources as res
from typing import Protocol, TypeAlias

import elasticai.creator.function_utils as F
from elasticai.creator import graph as g
from elasticai.creator import ir
from elasticai.creator import plugin as pl
from elasticai.creator.ir import Edge as Edge


class Node(ir.Node):
    implementation: str


Code: TypeAlias = tuple[str, Sequence[str]]


@dataclass
class PluginSpec(pl.PluginSpec):
    generated: tuple[str, ...]
    static_files: tuple[str, ...]


class Implementation(ir.Implementation[Node, ir.Edge]):
    type: str

    def __init__(self, *, graph: g.Graph[str], data: dict[str, ir.Attribute]):
        super().__init__(graph=graph, data=data)


class Ir2Verilog(ir.LoweringPass[Implementation, Code]):
    def __init__(self) -> None:
        super().__init__()
        self.__static_files: dict[str, Callable[[], str]] = {}
        self._loader = PluginLoader(self)

    def register_static(self, name: str, fn: Callable[[], str]) -> None:
        self.__static_files[name] = fn

    def __call__(self, args: Iterable[Implementation]) -> Iterator[Code]:
        for name, content in super().__call__(args):
            yield f"{name}.v", content
        for name, fn in self.__static_files.items():
            yield name, fn()

    def load_from_package(self, package: str) -> None:
        self._loader.load_from_package(package)


class PluginSymbol(pl.PluginSymbol[Ir2Verilog], Protocol):
    pass


class PluginLoader(pl.PluginLoader[Ir2Verilog]):
    """PluginLoader for Ir2Verilog passes."""

    def __init__(self, lowering: Ir2Verilog):
        builder: pl.SymbolFetcherBuilder[PluginSpec, Ir2Verilog] = (
            pl.SymbolFetcherBuilder(PluginSpec)
        )
        fetcher: pl.SymbolFetcher[Ir2Verilog] = (
            builder.add_fn(self.__get_generated)
            .add_fn(_StaticFile.make_symbols)
            .build()
        )
        super().__init__(
            fetch=fetcher,
            plugin_receiver=lowering,
        )

    def load_from_package(self, package: str) -> None:
        if "." not in package:
            package = f"elasticai.creator_plugins.{package}"
        super().load_from_package(package)

    @staticmethod
    def __get_generated(plugin: PluginSpec) -> Iterator[PluginSymbol]:
        if plugin.target_runtime == "verilog":
            yield from pl.import_symbols(plugin.package, plugin.generated)


class _StaticFile(PluginSymbol):
    _subfolder = "verilog"

    def __init__(self, name: str, package: str):
        self._name = name
        self._package = package

    @property
    def name(self) -> str:
        return self._name

    def load_into(self, receiver: Ir2Verilog):
        receiver.register_static(self.name, self)

    @classmethod
    def make_symbols(cls, p: PluginSpec) -> Iterator[pl.PluginSymbol[Ir2Verilog]]:
        if p.target_runtime == cls._subfolder:
            for name in p.static_files:
                yield cls(name=name, package=p.package)

    def __call__(self) -> str:
        file = res.files(self._package).joinpath(f"{self._subfolder}/{self.name}")
        return file.read_text()


TypeHandlerFn: TypeAlias = Callable[[Implementation], Code]


def _type_handler(
    name: str, fn: TypeHandlerFn
) -> pl.PluginSymbolFn[Ir2Verilog, [Implementation], Code]:
    def load_into(lower: Ir2Verilog) -> None:
        lower.register(name)(fn)

    return pl.make_plugin_symbol(load_into, fn)


def _type_handler_for_iterable(
    name: str, fn: Callable[[Implementation], Iterable[Code]]
) -> pl.PluginSymbolFn[Ir2Verilog, [Implementation], Iterable[Code]]:
    def load_into(lower: Ir2Verilog) -> None:
        lower.register_iterable(name)(fn)

    return pl.make_plugin_symbol(load_into, fn)


type_handler = F.FunctionDecorator(_type_handler)
type_handler_iterable = F.FunctionDecorator(_type_handler_for_iterable)



from abc import abstractmethod
from collections.abc import Collection, Hashable, Iterator, Mapping
from typing import Protocol, Self


class ReadOnlyGraph[T: Hashable](Protocol):
    @property
    @abstractmethod
    def nodes(self) -> Collection[T]: ...

    @abstractmethod
    def iter_edges(self) -> Iterator[tuple[T, T]]: ...

    @property
    @abstractmethod
    def predecessors(self) -> Mapping[T, Collection[T]]: ...

    @property
    @abstractmethod
    def successors(self) -> Mapping[T, Collection[T]]: ...


class Graph[T: Hashable](ReadOnlyGraph[T], Protocol):
    @abstractmethod
    def add_node(self, node: T) -> Self: ...

    @abstractmethod
    def add_edge(self, src: T, dst: T) -> Self: ...

    @abstractmethod
    def new(self) -> "Graph[T]": ...

    @abstractmethod
    def copy(self) -> "Graph[T]": ...



"""
The skeleton id should be computed and set after all other code is generated.
"""

import logging
from collections.abc import Iterable
from pathlib import Path
from warnings import warn

from elasticai.creator.hw_function_id import _HwFunctionIdHash
from elasticai.creator.vhdl.hw_function_id import _replace_id_in_vhdl


def compute_skeleton_id_hash(files: Iterable[Path]) -> bytes:
    warn(
        "this function will be deprecated soon, use `HwFunctionIdUpdater` instead.",
        category=DeprecationWarning,
    )
    logger = logging.getLogger(__name__)
    hash = _HwFunctionIdHash()

    for vhd_file in files:
        logger.debug(f"hashing {vhd_file.as_uri}")
        hash.update(vhd_file)
    digest = hash.digest()
    logger.debug(f"raw_digest is {digest.hex()}")

    return digest


def replace_id_in_vhdl(code: Iterable[str], id: bytes) -> Iterable[str]:
    warn(
        "this function will be deprecated soon, use `HwFunctionIdUpdater` instead.",
        category=DeprecationWarning,
    )
    """
    Look for a line that starts with `constant SKELETON_ID` and replace it with
    the given id.
    """

    return _replace_id_in_vhdl(code, id)


def update_skeleton_id_in_build_dir(build_dir: Path) -> bytes:
    warn(
        "this function will be deprecated soon, use `HwFunctionIdUpdater` instead.",
        category=DeprecationWarning,
    )
    """insert the id into the skeleton_pkg.vhd file under `build_dir`."""

    logger = logging.getLogger(__name__)
    logger.debug("updating skeleton id")
    skeleton_pkg = None

    def is_not_skeleton_pkg(f: Path) -> bool:
        nonlocal skeleton_pkg
        if f.name == "skeleton_pkg.vhd":
            skeleton_pkg = f
            return False
        return True

    def files_recursive():
        for f in build_dir.glob("**/*"):
            if f.is_file():
                yield f

    files_to_hash = filter(is_not_skeleton_pkg, files_recursive())
    id = compute_skeleton_id_hash(files_to_hash)
    logger.debug(f"computed id is {id!r}")
    if skeleton_pkg is None:
        raise IOError("skeleton_pkg.vhd not found in build folder")
    with open(skeleton_pkg, "r") as f:
        code: Iterable[str] = f.readlines()
    logger.debug("updating skeleton_pkg.vhd")
    code = replace_id_in_vhdl(code, id)
    with open(skeleton_pkg, "w") as f:
        for line in code:
            f.write(line)
            f.write("\n")
    logger.debug("done")
    return id



from elasticai.creator.ir2vhdl import Implementation
from elasticai.creator_plugins.lutron.lutron import lutron


def test_can_generate_vhdl_for_lutron():
    impl = Implementation(
        name="lut_0",
        type="lutron",
        data={
            "input_size": 2,
            "output_size": 2,
            "truth_table": (
                ("00", "11"),
                ("01", "10"),
                ("10", "01"),
                ("11", "11"),
            ),
        },
    )
    _, lines = lutron(impl)
    vhdl = tuple(map(str.strip, lines))
    expected = (
        "library ieee;",
        "use ieee.std_logic_1164.all;",
        "entity lut_0 is",
        "port (",
        "signal d_in : in std_logic_vector(2 - 1 downto 0);",
        "signal d_out : out std_logic_vector(2 - 1 downto 0)",
        ");",
        "end entity;",
        "",
        "architecture rtl of lut_0 is",
        "begin",
        "process (d_in) is",
        "begin",
        "case d_in is",
        'when b"00" => d_out <= b"11";',
        'when b"01" => d_out <= b"10";',
        'when b"10" => d_out <= b"01";',
        'when b"11" => d_out <= b"11";',
        "when others => d_out <= (others => 'X');",
        "end case;",
        "end process;",
        "end architecture;",
    )
    assert vhdl == expected



__all__ = [
    "Node",
    "Edge",
    "Implementation",
    "Lowerable",
    "LoweringPass",
    "Implementation",
    "VhdlNode",
    "edge",
    "vhdl_node",
    "FilterParameters",
    "Shape",
]
from elasticai.creator.ir import Lowerable, LoweringPass, Node
from elasticai.creator.ir2vhdl import (
    Edge,
    Implementation,
    Shape,
    VhdlNode,
    edge,
    vhdl_node,
)
from elasticai.creator_plugins.grouped_filter import FilterParameters



from typing import Protocol

from elasticai.creator.file_generation.savable import Path
from elasticai.creator.vhdl.design.design import Design
from elasticai.creator.vhdl.system_integrations.env5_constraints.env5_constraints import (
    ENV5Constraints,
)
from elasticai.creator.vhdl.system_integrations.middleware.middleware import Middleware
from elasticai.creator.vhdl.system_integrations.skeleton.skeleton import (
    LSTMSkeleton,
    Skeleton,
)
from elasticai.creator.vhdl.system_integrations.top.env5_reconfig_top import (
    ENV5ReconfigTop,
)


class SkeletonType(Protocol):
    def save_to(self, destination: Path) -> None: ...


class _FirmwareENv5Base:
    def __init__(self, skeleton: SkeletonType) -> None:
        self._skeleton = skeleton

    def save_to(self, destination: Path) -> None:
        def save_srcs(destination: Path):
            self._skeleton.save_to(destination.create_subpath("skeleton"))

            middleware = Middleware()
            middleware.save_to(destination.create_subpath("middleware"))

            env5_reconfig_top = ENV5ReconfigTop()
            env5_reconfig_top.save_to(destination.create_subpath("env5_reconfig_top"))

        def save_constraints(destination: Path):
            env5_config = ENV5Constraints()
            env5_config.save_to(destination.create_subpath("env5_config"))

        srcs = destination.create_subpath("srcs")
        constraints = destination.create_subpath("constraints")
        save_constraints(constraints)
        save_srcs(srcs)


class FirmwareENv5(_FirmwareENv5Base):
    def __init__(
        self,
        network: Design,
        x_num_values: int,
        y_num_values: int,
        id: list[int] | int,
        skeleton_version: str = "v1",
    ) -> None:
        super().__init__(
            skeleton=Skeleton(
                network_name=network.name,
                port=network.port,
                x_num_values=x_num_values,
                y_num_values=y_num_values,
                id=id,
                skeleton_version=skeleton_version,
            )
        )


class LSTMFirmwareENv5(_FirmwareENv5Base):
    def __init__(self, network: Design):
        super().__init__(skeleton=LSTMSkeleton(network.name))



import json
from pathlib import Path

from elasticai.creator.file_generation import find_project_root


def build_report_folder_and_testdata(dut_name: str, testdata: dict) -> Path:
    """Building the test/simulation folder which contains the test data and hardware design for testing in cocotb
    :param dut_name:        The name of the Top Module
    :param testdata:        Dictionary with test data/params data
    :return:                Path to the report folder containing hardware design and testpattern data
    """
    build_dir = find_project_root() / "build_test" / dut_name.lower()
    build_dir.mkdir(exist_ok=True, parents=True)

    file_name = "testdata.json"
    with open(build_dir / file_name, "w") as f0:
        json.dump(testdata, f0, indent=1)

    return build_dir


def read_testdata(dut_name: str) -> dict:
    """Reading the data as testpattern in the cocotb testbench
    :param dut_name:        The name of the Top Module DUT in the cocotb testbench (using dut._name)
    :return:                Dictionary with testpattern for testing the DUT
    """
    path_to_file = find_project_root() / "build_test" / dut_name.lower()
    file_name = "testdata.json".lower()
    with open(
        path_to_file / file_name,
        "r",
    ) as f:
        data = json.load(f)
    return data



from .conv1d import BatchNormedConv1d as BatchNormedConv1d
from .conv1d import Conv1d as Conv1d
from .hard_sigmoid import HardSigmoid as HardSigmoid
from .hard_tanh import HardTanh as HardTanh
from .linear import BatchNormedLinear as BatchNormedLinear
from .linear import Linear as Linear
from .math_operations import MathOperations as MathOperations
from .precomputed import AdaptableSiLU as AdaptableSiLU
from .precomputed import PReLU as PReLU
from .precomputed import Sigmoid as Sigmoid
from .precomputed import SiLU as SiLU
from .precomputed import Tanh as Tanh
from .relu import ReLU as ReLU



from .core import Edge, Node, edge, node
from .implementation import Graph, Implementation
from .lowering import Lowerable, LoweringPass

__all__ = [
    "Edge",
    "Node",
    "edge",
    "node",
    "LoweringPass",
    "Lowerable",
    "Implementation",
    "Graph",
]



from importlib import resources
from logging import getLogger
from pathlib import Path

from vunit import VUnit


def run_vunit_vhdl_testbenches(deps: list[str], test_dir: Path | str):
    logger = getLogger(__name__)
    if isinstance(test_dir, str):
        test_dir = Path(test_dir)
    test_dir = test_dir.absolute()
    logger.info("using test_dir {}".format(test_dir))
    vu = VUnit.from_argv()
    vu.add_vhdl_builtins()
    lib = vu.add_library("lib")
    for testbench in test_dir.glob("*_tb.vhd"):
        logger.info("adding testbench {}".format(testbench))
        lib.add_source_file(testbench.absolute())

    for dep in deps:
        vhdl_dir = resources.files(dep).joinpath("vhdl")
        if not vhdl_dir.is_dir():
            raise IOError("could not find directory under {}".format(vhdl_dir))
        for file in vhdl_dir.glob("*.vhd"):  # type: ignore
            lib.add_source_files(str(file.absolute()))
    vu.main()



from elasticai.creator_plugins.quantized_grads.base_modules import BatchNorm2d
from elasticai.creator_plugins.quantized_grads.fixed_point import (
    FixedPointConfigV2,
    QuantizeForwHTE,
    QuantizeParamToFixedPointHTE,
    QuantizeParamToFixedPointStochastic,
)


def test_batchnorm_fxp_init():
    conf = FixedPointConfigV2(8, 3)

    BatchNorm2d(
        math_ops=QuantizeForwHTE(forward_conf=conf),
        weight_quantization=QuantizeParamToFixedPointHTE(conf),
        bias_quantization=QuantizeParamToFixedPointStochastic(conf),
        num_features=3,
    )



from typing import Callable

import torch
from torch import Tensor

from .quantize_to_fixed_point import quantize_to_fxp_hte, quantize_to_fxp_stochastic


def get_autograd_func(
    forw_func: Callable[[Tensor, Tensor, Tensor, Tensor], Tensor],
    backw_func: Callable[[Tensor, Tensor, Tensor, Tensor], Tensor],
) -> (tuple)[type[torch.autograd.Function], type[torch.autograd.Function]]:
    class FixedPointForw(torch.autograd.Function):
        @staticmethod
        def forward(
            ctx,
            x: Tensor,
            forw_resolution_per_int: Tensor,
            forw_minimum_as_rational: Tensor,
            forw_maximum_as_rational: Tensor,
            *args,
            **kwargs,
        ):
            return forw_func(
                x,
                forw_resolution_per_int,
                forw_minimum_as_rational,
                forw_maximum_as_rational,
            )

        @staticmethod
        def backward(ctx, *grad_output):
            return *grad_output, None, None, None

    class FixedPointForwBackw(torch.autograd.Function):
        @staticmethod
        def forward(
            ctx,
            x: Tensor,
            forw_resolution_per_int: Tensor,
            forw_minimum_as_rational: Tensor,
            forw_maximum_as_rational: Tensor,
            backw_resolution_per_int: Tensor,
            backw_minimum_as_rational: Tensor,
            backw_maximum_as_rational: Tensor,
            *args,
            **kwargs,
        ):
            ctx.save_for_backward(
                backw_resolution_per_int,
                backw_minimum_as_rational,
                backw_maximum_as_rational,
            )
            return forw_func(
                x,
                forw_resolution_per_int,
                forw_minimum_as_rational,
                forw_maximum_as_rational,
            )

        @staticmethod
        def backward(ctx, *grad_output):
            (
                backw_resolution_per_int,
                backw_minimum_as_rational,
                backw_maximum_as_rational,
            ) = ctx.saved_tensors
            return (
                backw_func(
                    *grad_output,
                    backw_resolution_per_int,
                    backw_minimum_as_rational,
                    backw_maximum_as_rational,
                ),
                None,
                None,
                None,
                None,
                None,
                None,
            )

    return FixedPointForw, FixedPointForwBackw


QuantizeForwHTEAutograd, QuantizeForwHTEBackwHTEAutograd = get_autograd_func(
    quantize_to_fxp_hte, quantize_to_fxp_hte
)
(QuantizeForwStochasticAutograd, QuantizeForwStochasticBackwStochasticAutograd) = (
    get_autograd_func(quantize_to_fxp_stochastic, quantize_to_fxp_stochastic)
)

_, QuantizeForwHTEBackwStochasticAutograd = get_autograd_func(
    quantize_to_fxp_hte, quantize_to_fxp_stochastic
)



from elasticai.creator.graph.name_generation import NameRegistry


def test_can_generate_unique_name():
    reg = NameRegistry()
    reg.prepopulate(["a", "b", "c"])
    assert reg.get_unique_name("a") == "a_1"
    assert reg.get_unique_name("a") == "a_2"


def test_name_is_not_altered_if_already_unique():
    reg = NameRegistry()
    reg.prepopulate(["a", "b", "c"])
    assert reg.get_unique_name("d") == "d"


def test_prepopulated_suffixes_are_counted():
    reg = NameRegistry()
    reg.prepopulate(["a", "a_1", "a_2"])
    assert reg.get_unique_name("a") == "a_3"


def test_start_counting_from_highest_suffix():
    reg = NameRegistry()
    reg.prepopulate(["a_3"])
    assert reg.get_unique_name("a") == "a_4"


def test_count_either_no_suffix_or_zero_as_one():
    reg = NameRegistry()
    reg.prepopulate(["a", "a_0"])
    assert reg.get_unique_name("a") == "a_1"


def test_use_maximum_of_suffixes():
    reg = NameRegistry()
    reg.prepopulate(["a", "a_3", "a_2"])

    assert reg.get_unique_name("a") == "a_4"



import pytest

from elasticai.creator.ir2vhdl import (
    Implementation,
    LoweringPass,
    Shape,
    edge,
    vhdl_node,
)
from elasticai.creator_plugins.grouped_filter import FilterParameters
from elasticai.creator_plugins.time_multiplexed_sequential.src import (
    network as _network_handler,
)
from elasticai.creator_plugins.time_multiplexed_sequential.src import (
    sequential,
)

from .high_level_network import network


@pytest.fixture
def lower():
    _lower = LoweringPass()
    _lower.register(sequential)
    _lower.register_iterable(_network_handler)
    return _lower


def test_sets_input_correctly_for_version_without_sliding_window(lower):
    lowered = tuple(
        lower(
            [
                network(
                    input_shape=(1, 3),
                    kernel_size=3,
                    out_channels=1,
                ),
            ]
        )
    )[0]
    expected_input_node = vhdl_node(
        name="input",
        type="input",
        implementation="",
        input_shape=(1, 3),
        output_shape=(1, 3),
    )
    assert lowered.nodes["input"] == expected_input_node


def test_first_shift_reg_has_three_output_wires():
    in_channels = 2
    length = 4
    _net = network(
        input_shape=(in_channels, length), kernel_size=2, out_channels=2, stride=1
    )

    lowered = sequential(_net)
    node = lowered.nodes["shift_register_i0"]
    assert node.output_shape == (2, 2)


def test_shift_reg_predecessor_is_conv1():
    in_channels = 2
    length = 4
    _net = network(
        input_shape=(in_channels, length), kernel_size=2, out_channels=2, stride=1
    )

    lowered = sequential(_net)
    p = {n.name for n in (iter(lowered.successors("shift_register_i0").values()))}

    assert p == {"conv1_i0"}


def test_second_conv_has_three_input_wires():
    kernel_size = 1
    _net = network(
        input_shape=(3, 2), kernel_size=kernel_size, out_channels=3, stride=1
    )
    hl_conv1 = _net.nodes["conv1_i0"]
    lowered = sequential(_net)
    low_level_conv1 = lowered.nodes["conv1_i0"]
    expected_shape = Shape(hl_conv1.input_shape.depth, kernel_size)
    assert low_level_conv1.input_shape == expected_shape


class TestsForSecondConv:
    def network(self):
        return network(input_shape=(3, 2), kernel_size=1, out_channels=3, stride=1)

    def get_hl_conv(self):
        pass

    def test_high_and_low_level_have_same_in_channels(self):
        _net = self.network()
        hl_conv1 = _net.nodes["conv1_i0"]
        lowered = sequential(_net)
        low_level_conv1 = lowered.nodes["conv1_i0"]
        assert low_level_conv1.input_shape.depth == hl_conv1.input_shape.depth


def test_add_correct_edges(lower):
    net = network(
        input_shape=(2, 7),
        kernel_size=3,
        out_channels=1,
    )
    lowered = tuple(lower([net]))[0]
    edges = tuple(lowered.edges.values())
    expected = (
        edge(
            src="input",
            dst="conv0_i0",
            src_dst_indices=tuple(),
        ),
        edge(
            src="conv0_i0",
            dst="striding_shift_register_i0",
            src_dst_indices=tuple(),
        ),
        edge(
            src="striding_shift_register_i0",
            dst="conv1_i0",
            src_dst_indices=tuple(),
        ),
        edge(src="conv1_i0", dst="output", src_dst_indices=tuple()),
    )
    assert edges == expected


def test_remember_stride_after_pointwise_filter() -> None:
    in_channels = 1
    input_shape = Shape(in_channels, 20)

    net = Implementation(name="network", type="network")
    net.add_nodes(
        (
            vhdl_node(
                name="input",
                input_shape=input_shape,
                output_shape=input_shape,
                type="input",
                implementation="",
            ),
            vhdl_node(
                name="f0_i0",
                type="filter",
                implementation="f0",
                input_shape=(1, 20),
                output_shape=(2, 10),
                attributes=dict(
                    filter_parameters=FilterParameters(
                        kernel_size=2, in_channels=1, out_channels=2, stride=2
                    ).as_dict()
                ),
            ),
            vhdl_node(
                name="f1_i0",
                type="filter",
                implementation="f1",
                input_shape=(2, 10),
                output_shape=(3, 10),
                attributes=dict(
                    filter_parameters=FilterParameters(
                        kernel_size=1, out_channels=3, in_channels=2
                    ).as_dict()
                ),
            ),
            vhdl_node(
                name="f2_i0",
                type="filter",
                implementation="f2",
                input_shape=(3, 10),
                output_shape=(2, 9),
                attributes=dict(
                    filter_parameters=FilterParameters(
                        kernel_size=2, out_channels=2, in_channels=3
                    ).as_dict()
                ),
            ),
            vhdl_node(
                name="output",
                type="output",
                implementation="",
                input_shape=(2, 9),
                output_shape=(2, 9),
            ),
        )
    )
    edge_sequence = ["input", "f0_i0", "f1_i0", "f2_i0", "output"]
    net.add_edges(
        (
            edge(src, dst, tuple())
            for src, dst in zip(edge_sequence[:-1], edge_sequence[1:])
        )
    )
    lowered = sequential(net)
    follower_of_pointwise = list(lowered.successors("f1_i0").values())[0]
    assert follower_of_pointwise.type == "striding_shift_register"



from elasticai.creator.ir2vhdl import Implementation, Shape, edge, vhdl_node

from ..clocked_combinatorial import clocked_combinatorial


def test_using_two_clocked_components_connects_their_valid_signals():
    impl = Implementation(name="test")
    impl.add_nodes(
        (
            vhdl_node(
                name="input",
                type="input",
                implementation="",
                input_shape=Shape(2, 4),
                output_shape=Shape(2, 4),
            ),
            vhdl_node(
                "sliding_window",
                type="sliding_window",
                implementation="sliding_window",
                input_shape=Shape(2, 4),
                output_shape=Shape(2, 2),
            ),
            vhdl_node(
                "conv",
                type="unclocked_combinatorial",
                implementation="conv",
                input_shape=Shape(2, 4),
                output_shape=Shape(1, 1),
            ),
            vhdl_node(
                "shift_register",
                type="shift_register",
                implementation="shift_register",
                input_shape=Shape(1, 1),
                output_shape=(1, 3),
            ),
            vhdl_node(
                "output",
                "output",
                "",
                input_shape=Shape(1, 3),
                output_shape=Shape(1, 3),
            ),
        )
    )
    impl.nodes["sliding_window"].data["generic_map"] = dict(stride=1)
    impl.add_edges(
        (
            edge("input", "sliding_window", []),
            edge("sliding_window", "conv", []),
            edge("conv", "shift_register", []),
            edge("shift_register", "output", []),
        )
    )
    result = clocked_combinatorial(impl)
    _, lines = result
    lines = set(l.lstrip() for l in lines)
    valid_ins = {line for line in lines if line.startswith("valid_in")}
    assert "valid_in_shift_register <= valid_out_sliding_window;" in valid_ins



from typing import TypeAlias

SizeT: TypeAlias = tuple[int] | tuple[int, int] | tuple[int, int, int]
Attribute: TypeAlias = (
    int
    | float
    | str
    | list["Attribute"]
    | dict[str, "Attribute"]
    | list[int]
    | tuple["Attribute", ...]
)



import csv
from pathlib import Path as PyPath
from typing import Any

from elasticai.creator.file_generation.savable import Path


class Testbench:
    @property
    def name(self) -> str:
        raise NotImplementedError

    def save_to(self, destination: Path) -> None:
        raise NotImplementedError

    def prepare_inputs(self, *args: Any, **kwargs: Any) -> Any:
        raise NotImplementedError

    def parse_reported_content(self, *args, **kwargs: Any) -> Any:
        raise NotImplementedError


class SimulatedLayer:
    def __init__(
        self, testbench: Testbench, simulator_constructor, working_dir: str | PyPath
    ):
        self._testbench = testbench
        self._simulator_constructor = simulator_constructor
        self._working_dir = (
            working_dir if isinstance(working_dir, PyPath) else PyPath(working_dir)
        )
        self._inputs_file_path = (
            self._working_dir / f"{self._testbench.name}_inputs.csv"
        )

    def __call__(self, inputs: Any) -> Any:
        runner = self._simulator_constructor(
            workdir=str(self._working_dir), top_design_name=self._testbench.name
        )
        inputs = self._testbench.prepare_inputs(inputs)
        self._write_csv(inputs)
        runner.add_generic(INPUTS_FILE_PATH=str(self._inputs_file_path.absolute()))
        runner.initialize()
        runner.run()
        actual = self._testbench.parse_reported_content(runner.getReportedContent())
        return actual

    def _write_csv(self, inputs):
        with self._inputs_file_path.open("w") as f:
            header = [x for x in inputs[0].keys()]
            writer = csv.DictWriter(
                f,
                fieldnames=header,
                lineterminator="\n",
                delimiter=" ",
            )
            writer.writeheader()
            writer.writerows(inputs)



from collections.abc import Callable

import pytest
import torch

from elasticai.creator.nn.fixed_point.conv1d.layer.conv1d import Conv1d


@pytest.fixture
def conv1d() -> Conv1d:
    return Conv1d(
        total_bits=16,
        frac_bits=8,
        in_channels=3,
        out_channels=4,
        kernel_size=2,
        bias=False,
        signal_length=5,
    )


@pytest.fixture
def build_5_total_1_frac_conv1d():
    def build(in_channels, out_channels, kernel_size, signal_length):
        return Conv1d(
            total_bits=5,
            frac_bits=1,
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            signal_length=signal_length,
        )

    return build


@pytest.fixture
def to_1d_input_tensor() -> Callable[[list], torch.Tensor]:
    def get_list_nesting_depth(l):
        level = 0
        while isinstance(l, list):
            level += 1
            if len(l) > 0:
                l = l[0]
            else:
                break
        return level

    def convert(data: list):
        correctly_nested = data
        actual_depth = get_list_nesting_depth(data)
        required_depth = 3
        for _ in range(required_depth - actual_depth):
            correctly_nested = [correctly_nested]
        return torch.tensor(correctly_nested)

    return convert


@pytest.fixture
def conv_with_4_total_1_frac_and_kernel_size_2() -> Conv1d:
    return Conv1d(
        total_bits=4,
        frac_bits=1,
        kernel_size=2,
        signal_length=2,
        bias=False,
        in_channels=1,
        out_channels=1,
    )


@pytest.fixture
def tensor_of_size_3_15() -> torch.Tensor:
    return torch.rand(3, 15)


@pytest.fixture
def batched_input(tensor_of_size_3_15):
    return tensor_of_size_3_15


@pytest.fixture
def tensor_of_size_3_5_7() -> torch.Tensor:
    return torch.rand(3, 5, 7)


def test_that_batch_dimension_is_kept(
    build_5_total_1_frac_conv1d, batched_input: torch.Tensor
) -> None:
    in_channels = 5
    signal_length = 7
    batch_size = 3
    kernel_size = 2
    out_channels = 9
    input_data = torch.rand(batch_size, in_channels, signal_length)
    conv1d = build_5_total_1_frac_conv1d(
        in_channels=in_channels,
        out_channels=out_channels,
        kernel_size=kernel_size,
        signal_length=signal_length,
    )
    prediction = conv1d(input_data)
    expected_size = (batch_size, out_channels, signal_length - kernel_size + 1)

    assert expected_size == prediction.size()


def test_that_batched_output_is_flat(
    conv1d: Conv1d, batched_input: torch.Tensor
) -> None:
    prediction = conv1d(batched_input)
    tensor_rank = prediction.dim()
    expected_rank = 2
    assert expected_rank == tensor_rank


@pytest.mark.parametrize(
    "signal_length, inputs, expected",
    [
        (2, [1.0, 1.0], [2.0]),
        (3, [1.0, 2.0, 4.0], [3.0, 6.0]),
    ],
)
def test_ones_weight_kernel_convolution(
    signal_length: int, inputs: list[float], expected: list[float], to_1d_input_tensor
) -> None:
    input_tensor = to_1d_input_tensor(inputs)
    conv = Conv1d(
        total_bits=16,
        frac_bits=8,
        in_channels=1,
        out_channels=1,
        bias=False,
        signal_length=signal_length,
        kernel_size=2,
    )
    conv.weight.data = torch.ones_like(conv.weight)
    prediction = conv(input_tensor)
    expected = to_1d_input_tensor(expected).tolist()
    assert expected == prediction.tolist()


def test_no_overflow_after_multiplication(
    conv_with_4_total_1_frac_and_kernel_size_2, to_1d_input_tensor
) -> None:
    conv = conv_with_4_total_1_frac_and_kernel_size_2
    conv.weight.data = 3.5 * torch.ones_like(conv.weight)
    prediction = conv(to_1d_input_tensor([3.5, -3]))
    expected = to_1d_input_tensor(
        [1.5]
    ).tolist()  # quantize(3.5 * 3.5 - 3 * 3.5) == quantize(1.75) == 1.5
    assert expected == prediction.tolist()


def test_no_underflow_after_multiplication(
    conv_with_4_total_1_frac_and_kernel_size_2, to_1d_input_tensor
) -> None:
    conv = conv_with_4_total_1_frac_and_kernel_size_2
    conv.weight.data = 0.5 * torch.ones_like(conv.weight)
    prediction = conv(to_1d_input_tensor([0.5, 0]))
    expected = to_1d_input_tensor(
        0
    ).tolist()  # quantize(0.5 * 0.5 + 0.5 * 0) == quantize(0.25) == 0
    assert expected == prediction.tolist()


@pytest.mark.parametrize(
    "inputs, expected",
    [([2.0, 2.0], [3.5]), ([-2.0, -3.0], [-4.0])],
)
def test_fxp_operations_additive_over_and_underflow(
    conv_with_4_total_1_frac_and_kernel_size_2,
    inputs: list[float],
    expected: list[float],
    to_1d_input_tensor,
) -> None:
    conv = conv_with_4_total_1_frac_and_kernel_size_2
    conv.weight.data = torch.ones_like(conv.weight)
    input_tensor = to_1d_input_tensor(inputs)
    prediction = conv(input_tensor)
    expected = to_1d_input_tensor(expected).tolist()
    assert expected == prediction.tolist()


def test_ensure_bias_is_0_5(to_1d_input_tensor) -> None:
    conv = Conv1d(
        total_bits=4,
        frac_bits=1,
        kernel_size=2,
        signal_length=2,
        bias=True,
        in_channels=1,
        out_channels=1,
    )
    conv.weight.data = torch.ones_like(conv.weight)
    conv.bias.data = 0.5 * torch.ones_like(conv.bias)  # type: ignore
    inputs = to_1d_input_tensor([0.0, 0.0])
    predictions = conv(inputs).tolist()
    expected = to_1d_input_tensor(0.5).tolist()
    assert expected == predictions



import subprocess
import tomllib
from pathlib import Path

import pytest
import serial
import torch
from elasticai.runtime.env5.usb import UserRemoteControl, get_env5_port

from elasticai.creator.arithmetic import (
    FxpArithmetic,
    FxpParams,
)
from elasticai.creator.file_generation.on_disk_path import OnDiskPath
from elasticai.creator.nn import Sequential
from elasticai.creator.nn.fixed_point import Conv1d
from elasticai.creator.vhdl.system_integrations.firmware_env5 import FirmwareENv5
from tests.system_tests.helper.parse_tensors_to_bytearray import (
    parse_bytearray_to_fxp_tensor,
    parse_fxp_tensor_to_bytearray,
)


def create_vhd_files(
    output_dir: str,
    num_inputs: int,
    num_outputs: int,
    num_in_channels: int,
    num_out_channels: int,
    kernel_size: int,
    total_bits: int,
    frac_bits: int,
    skeleton_id: list[int],
) -> Sequential:
    nn = Sequential(
        Conv1d(
            total_bits=total_bits,
            frac_bits=frac_bits,
            in_channels=num_in_channels,
            out_channels=num_out_channels,
            signal_length=num_inputs,
            kernel_size=kernel_size,
        )
    )
    nn[0].weight.data = torch.ones_like(nn[0].weight) * 2
    nn[0].bias.data = torch.ones_like(nn[0].bias) * -1
    destination = OnDiskPath(output_dir)
    my_design = nn.create_design("nn")
    my_design.save_to(destination.create_subpath("srcs"))

    firmware = FirmwareENv5(
        network=my_design,
        x_num_values=num_inputs,
        y_num_values=num_outputs,
        id=skeleton_id,
        skeleton_version="v2",
    )
    firmware.save_to(destination)
    return nn


def vivado_build_binfile(input_dir: str, binfile_dir: str):
    print(f"Building binfile in {binfile_dir}")
    with open("./tests/system_tests/vivado_build_server_conf.toml", "rb") as f:
        config = tomllib.load(f)
    out = subprocess.run(
        [
            "bash",
            "./utils/autobuild_binfile_vivado2021.1.sh",
            config["username"],
            config["ip"],
            input_dir,
            binfile_dir,
        ],
        capture_output=True,
    )
    print(f"{out.stdout=}")


@pytest.mark.hardware
def test_conv1d_runtime():
    # dev_address = "COM8"
    dev_address = get_env5_port()
    vhdl_dir = "./tests/system_tests/conv1d/build_dir"
    binfile_dir = "./tests/system_tests/conv1d/build_dir_output"
    binfile_path = Path(binfile_dir + "/output/env5_top_reconfig.bin")
    skeleton_id = list(range(15, -1, -1))
    skeleton_id_as_bytearray = bytearray()
    for x in skeleton_id:
        skeleton_id_as_bytearray.extend(
            x.to_bytes(length=1, byteorder="little", signed=False)
        )

    total_bits = 8
    frac_bits = 2
    num_inputs = 5
    kernel_size = 3
    num_in_channels = 1
    num_out_channels = 1
    # batches = 2
    num_outputs = num_inputs - kernel_size + 1
    nn = create_vhd_files(
        output_dir=vhdl_dir,
        num_inputs=num_inputs,
        num_outputs=num_outputs,
        num_in_channels=num_in_channels,
        num_out_channels=num_out_channels,
        kernel_size=kernel_size,
        total_bits=total_bits,
        frac_bits=frac_bits,
        skeleton_id=skeleton_id,
    )

    fxp_conf = FxpArithmetic(
        FxpParams(total_bits=total_bits, frac_bits=frac_bits, signed=True)
    )

    # inputs = fxp_conf.as_rational(fxp_conf.as_integer(torch.rand(batches, num_in_channels, num_inputs)))
    inputs = fxp_conf.as_rational(
        fxp_conf.cut_as_integer(
            torch.Tensor(
                [
                    [[0.0, 0.0, 0.0, 0.0, 0.0]],
                    [[1.0, 0.0, 0.0, 0.0, 0.0]],
                    [[0.0, 1.0, 0.0, 0.0, 0.0]],
                    [[0.0, 0.0, 1.0, 0.0, 0.0]],
                    [[0.0, 0.0, 0.0, 1.0, 0.0]],
                    [[0.0, 0.0, 0.0, 0.0, 1.0]],
                    [[2.0, 0.0, 0.0, 0.0, 0.0]],
                    [[1.0, 1.0, 0.0, 0.0, 0.0]],
                    [[1.0, 0.0, 1.0, 0.0, 0.0]],
                    [[1.0, 0.0, 0.0, 1.0, 0.0]],
                    [[1.0, 0.0, 0.0, 0.0, 1.0]],
                    [[3.0, 0.0, 0.0, 0.0, 0.0]],
                    [[1.0, 2.0, 0.0, 0.0, 0.0]],
                    [[1.0, 0.0, 2.0, 0.0, 0.0]],
                    [[1.0, 0.0, 0.0, 2.0, 0.0]],
                    [[1.0, 0.0, 0.0, 0.0, 2.0]],
                    [[1.0, 1.0, 0.0, 0.0, 0.0]],
                    [[0.0, 1.0, 1.0, 0.0, 0.0]],
                    [[0.0, 0.0, 1.0, 1.0, 0.0]],
                    [[0.0, 0.0, 0.0, 1.0, 1.0]],
                    [[-1.0, 0.0, 0.0, 0.0, 0.0]],
                    [[0.0, -1.0, 0.0, 0.0, 0.0]],
                    [[0.0, 0.0, -1.0, 0.0, 0.0]],
                    [[0.0, 0.0, 0.0, -1.0, 0.0]],
                    [[0.0, 0.0, 0.0, 0.0, -1.0]],
                    [[2.0, 0.0, 0.0, 0.0, 0.0]],
                    [[0.0, 2.0, 0.0, 0.0, 0.0]],
                    [[0.0, 0.0, 2.0, 0.0, 0.0]],
                    [[0.0, 0.0, 0.0, 2.0, 0.0]],
                    [[0.0, 0.0, 0.0, 0.0, 2.0]],
                ]
            )
        )
    )

    expected_outputs = nn(inputs)
    vivado_build_binfile(vhdl_dir, binfile_dir)

    with serial.Serial(dev_address) as serial_con:
        flash_start_address = 0

        urc = UserRemoteControl(device=serial_con)
        urc.send_and_deploy_model(
            binfile_path, flash_start_address, skeleton_id_as_bytearray
        )
        # urc.deploy_model(flash_start_address, skeleton_id_as_bytearray)

        batch_data = parse_fxp_tensor_to_bytearray(inputs, total_bits, frac_bits)
        inference_result = list()
        for i, sample in enumerate(batch_data):
            batch_result = urc.inference_with_data(
                sample, num_outputs * num_out_channels
            )
            my_result = parse_bytearray_to_fxp_tensor(
                [batch_result],
                total_bits,
                frac_bits,
                (1, num_out_channels, num_outputs),
            )
            print(
                f"Batch {i}: {my_result} == {expected_outputs[i]}, for input"
                f" {inputs[i]}"
            )

            inference_result.append(batch_result)
        actual_result = parse_bytearray_to_fxp_tensor(
            inference_result, total_bits, frac_bits, expected_outputs.shape
        )

        assert torch.equal(actual_result, expected_outputs)



from collections.abc import Callable, Iterable, Iterator, Mapping
from typing import (
    Any,
    Generic,
    MutableMapping,
    ParamSpec,
    Protocol,
    Self,
    TypeVar,
    cast,
    overload,
)

from elasticai.creator.graph import Graph
from elasticai.creator.ir.base import Attribute, IrData, read_only_field

from .core import Edge, Node

N = TypeVar("N", bound=Node, covariant=True)
E = TypeVar("E", bound=Edge, covariant=True)

StoredT = TypeVar("StoredT", bound=Attribute)
VisibleT = TypeVar("VisibleT")
P = ParamSpec("P")


class NodeFn(Protocol[N]):
    def __call__(self, name: str, data: dict[str, Attribute]) -> N: ...


class EdgeFn(Protocol[E]):
    def __call__(self, src: str, dst: str, data: dict[str, Attribute]) -> E: ...


class Implementation(IrData, Generic[N, E], create_init=False):
    __slots__ = (
        "data",
        "graph",
        "_node_fn",
        "_edge_fn",
        "_nodes",
        "_edges",
    )

    name: str
    type: str

    @overload
    def __init__(
        self: "Implementation[N, E]",
        *,
        graph: Graph[str],
        edge_fn: EdgeFn[E],
        node_fn: NodeFn[N],
        data: dict[str, Attribute] | None = None,
    ) -> None: ...

    @overload
    def __init__(
        self: "Implementation[N, Edge]",
        *,
        graph: Graph,
        node_fn: NodeFn[N],
        data: dict[str, Attribute] | None = None,
    ) -> None: ...

    @overload
    def __init__(
        self: "Implementation[Node, E]",
        *,
        graph: Graph[str],
        edge_fn: EdgeFn[E],
        data: dict[str, Attribute] | None = None,
    ) -> None: ...

    @overload
    def __init__(
        self: "Implementation[Node, Edge]",
        *,
        graph: Graph[str],
        data: dict[str, Attribute] | None = None,
    ) -> None: ...

    def __init__(
        self,
        *,
        graph: Graph[str],
        node_fn: NodeFn = Node,
        edge_fn: EdgeFn = Edge,
        data: dict[str, Attribute] | None = None,
    ) -> None:
        """Create a new Implementation. Nodes and edges from `data` will not be automatically added to the graph.

        The constructor will combine `graph` and `data` so we can access nodes and edges of type `N` and `E` respectively.
        And use these to manipulate the underlying graph structure.
        It is the callers responsibility to keep `graph` and `data` in sync.
        This might seem like a limitation at first, but in fact it is a feature.
        It allows us to combine existing graph structures with new underlying data.
        A common use case for this is to handle subgraphs as in the following example:

        ```python
        subgraph = Graph()

        for node in original_impl.nodes.values():
            if fulfills_constraint(node):
                subgraph.add_node(node.name)

        for src, dst in original_impl.edges:
            if fulfills_constraint(original_impl.nodes[src]) and fulfills_constraint(original_impl.nodes[dst]):
                subgraph.add_edge(src, dst)

        new_impl = Implementation(graph=subgraph, data=original_impl.data)

        for node in new_impl.nodes.values():  # now we can iterate over the nodes of the subgraph
            do_something(node)
        ```
        """
        if data is None:
            data = {"nodes": {}, "edges": {}}
        if "nodes" not in data:
            data["nodes"] = {}
        if "edges" not in data:
            data["edges"] = {}
        super().__init__(data)

        self.graph = graph
        self._nodes: dict[str, Attribute] = cast(
            dict[str, Attribute], self.data["nodes"]
        )
        self._edges: dict[str, dict[str, Attribute]] = cast(
            dict[str, dict[str, Attribute]], self.data["edges"]
        )

        self._node_fn = node_fn
        self._edge_fn = edge_fn

    @property
    def _node_data(self) -> dict[str, Attribute]:
        return cast(dict[str, Attribute], self.data["nodes"])

    @property
    def _edge_data(self) -> MutableMapping[tuple[str, str], Attribute]:
        return _NestedDictToTupleKeyAdapter(
            cast(dict[str, dict[str, "Attribute"]], self.data["edges"])
        )

    @overload
    def add_node(self, n: Node) -> Self: ...

    @overload
    def add_node(self, *, name: str, data: dict[str, Attribute]) -> Self: ...

    def add_node(self, *args, **kwargs) -> Self:
        if len(args) + len(kwargs) == 1:
            bound = _bind_args(["node"], {}, *args, **kwargs)
            node = bound["node"]
            return self._add_node(node.name, node.data)
        else:
            bound = _bind_args(["name", "data"], {"data": {}}, *args, **kwargs)
            return self._add_node(bound["name"], bound["data"])

    def _add_node(self, name: str, data: dict[str, Attribute]) -> Self:
        self.graph.add_node(name)
        self._node_data[name] = data
        return self

    @overload
    def add_edge(self, e: Edge) -> Self: ...

    @overload
    def add_edge(self, src: str, dst: str, data: dict[str, Attribute]) -> Self: ...

    def add_edge(self, *args, **kwargs) -> Self:
        if len(args) + len(kwargs) == 1:
            if len(args) == 1:
                e = args[0]
            else:
                e = kwargs["edge"]
            return self.add_edge(e.src, e.dst, e.data)
        else:
            bound = _bind_args(["src", "dst", "data"], {"data": {}}, *args, **kwargs)
            self.graph.add_edge(bound["src"], bound["dst"])
            self._edge_data[(bound["src"], bound["dst"])] = bound["data"]
            return self

    def add_nodes(self, ns: Iterable[Node]) -> Self:
        for n in ns:
            self.add_node(n)
        return self

    def add_edges(self, es: Iterable[Edge]) -> Self:
        for e in es:
            self.add_edge(e)
        return self

    def successors(self, node: str | Node) -> Mapping[str, N]:
        if not isinstance(node, str):
            node = node.name

        def _successors():
            return self.graph.successors[node]

        return self.get_node_mapping(_successors)

    def predecessors(self, node: str | Node) -> Mapping[str, N]:
        if not isinstance(node, str):
            node = node.name

        def _predecessors():
            return self.graph.predecessors[node]

        return self.get_node_mapping(_predecessors)

    def get_node_mapping(self, keys: Callable[[], Iterable[str]]) -> Mapping[str, N]:
        """Create a read-only mapping of names to Nodes in the order of `keys()`."""
        return _ReadOnlyMappingInOrderAsIterable(keys, self._node_data, self._node_fn)

    def get_edge_mapping(
        self, keys: Callable[[], Iterable[tuple[str, str]]]
    ) -> Mapping[tuple[str, str], E]:
        """Create a read-only mapping of (src, dst) pairs to Edges in the order of `keys()`."""
        return _ReadOnlyMappingInOrderAsIterable(
            keys, self._edge_data, self._construct_edge
        )

    def _construct_edge(
        self, src_dst: tuple[str, str], data: dict[str, Attribute]
    ) -> E:
        src, dst = src_dst
        return self._edge_fn(src, dst, data)

    @read_only_field
    def nodes(self, _: dict[str, Attribute]) -> Mapping[str, N]:
        return self.get_node_mapping(lambda: self.graph.nodes)

    @read_only_field
    def edges(self, _: dict[str, Attribute]) -> Mapping[tuple[str, str], E]:
        return self.get_edge_mapping(self.graph.iter_edges)

    def as_dict(self) -> dict[str, Attribute]:
        data = self.data.copy()
        edges: dict[str, dict[str, Attribute]] = {}
        for (src, dst), d in self._edge_data.items():
            if src in edges:
                edges[src][dst] = d
            else:
                edges[src] = {dst: d}
        data["edges"] = cast(Attribute, edges)

        return data

    def sync_data_with_graph(self) -> Self:
        """Removes nodes/edges from data that are not in the graph, add empty fields for new nodes/edges."""
        nodes_to_remove = set()
        for n in self._nodes:
            if n not in self.graph.nodes:
                nodes_to_remove.add(n)
        for n in nodes_to_remove:
            del self._nodes[n]
        for n in self.graph.nodes:
            if n not in self._nodes:
                self._nodes[n] = {}

        edges_to_remove = set()
        edges_to_keep = set(self.graph.iter_edges())
        for src in self._edges:
            for dst in self._edges[src]:
                if (src, dst) not in edges_to_keep:
                    edges_to_remove.add((src, dst))

        for src, sink in edges_to_remove:
            if src in self._edges:
                del self._edges[src][sink]
            if len(self._edges[src]) == 0:
                del self._edges[src]

        for src, sink in edges_to_keep:
            if src not in self._edges:
                self._edges[src] = {}
            if sink not in self._edges[src]:
                self._edges[src][sink] = {}

        return self

    def load_from_dict(
        self,
        d: dict[str, Any],
    ) -> "Implementation[N, E]":
        """Load attributes, nodes and edges from a dictionary, that was created by `as_dict`.

        Opposed to the constructor, this will add all nodes and edges to the underlying graph.
        :::{important}
        This will override all state of the current object.
        :::
        """
        self.data = d.copy()
        g = self.graph
        self.graph = self.graph.new()
        del g
        for n in self._node_data.keys():
            self.graph.add_node(n)
        for src, dst in self._edge_data.keys():
            self.graph.add_edge(src, dst)

        return self


_K = TypeVar("_K")
_V = TypeVar("_V")
_K2 = TypeVar("_K2")


class _NestedDictToTupleKeyAdapter(MutableMapping[tuple[_K, _K2], _V]):
    def __init__(self, wrapped: dict[_K, dict[_K2, _V]]):
        self.wrapped = wrapped

    def __getitem__(self, k: tuple[_K, _K2]) -> _V:
        return self.wrapped[k[0]][k[1]]

    def __len__(self) -> int:
        acc = 0
        for v in self.wrapped.values():
            acc += len(v)
        return acc

    def __contains__(self, k: object) -> bool:
        if not isinstance(k, tuple) or len(k) != 2:
            return False
        return k[0] in self.wrapped and k[1] in self.wrapped[k[0]]

    def __delitem__(self, key):
        self.wrapped(key[0]).__delitem__(key[1])

    def __iter__(self) -> Iterator[tuple[_K, _K2]]:
        for k0 in self.wrapped:
            for k1 in self.wrapped[k0]:
                yield k0, k1

    def __setitem__(self, k: tuple[_K, _K2], v: _V) -> None:
        if k[0] not in self.wrapped:
            self.wrapped[k[0]] = {}

        self.wrapped[k[0]][k[1]] = v


class _ReadOnlyMappingInOrderAsIterable(Mapping[_K, _V]):
    def __init__(
        self,
        iterable: Callable[[], Iterable[_K]],
        d: Mapping[_K, Any],
        value_constructor: Callable[[_K, Any], _V],
    ):
        self._iterable = iterable
        self._d = d
        self._value_constructor = value_constructor

    def __iter__(self) -> Iterator[_K]:
        yield from self._iterable()

    def __len__(self) -> int:
        return len(self._d)

    def __contains__(self, k: object) -> bool:
        return k in self._d

    def __getitem__(self, k: _K) -> _V:
        return self._value_constructor(k, self._d[k])

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Mapping):
            return NotImplemented
        return dict(self) == dict(other)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({dict(self)})"


def _bind_args(keywords: list[str], optionals: dict[str, Any], *args, **kwargs) -> dict:
    bound = dict()
    keys = keywords.copy()
    opts = optionals.copy()
    for key, arg in zip(keys, args):
        bound[key] = arg
    for key, value in kwargs.items():
        if key in bound:
            raise TypeError("failed to bind arguments: multiple values for argument")
        bound[key] = value
    for key, value in opts.items():
        if key not in bound:
            bound[key] = value

    missing_args = set()
    for k in keywords:
        if k not in bound:
            missing_args.add(k)
    if len(missing_args) > 0:
        raise TypeError(f"failed to bind arguments: missing argument {missing_args}")

    return bound



__all__ = [
    "Implementation",
    "VhdlNode",
    "Edge",
    "FilterParameters",
    "edge",
    "vhdl_node",
]

from elasticai.creator.ir.helpers import FilterParameters

from elasticai.creator.ir2vhdl import Edge, Implementation, VhdlNode, edge, vhdl_node



from collections.abc import Hashable, Iterable, Iterator
from typing import Protocol, TypeVar

HashableT = TypeVar("HashableT", bound=Hashable)


class NodeNeighbourFn(Protocol[HashableT]):
    def __call__(self, node: HashableT) -> Iterable[HashableT]: ...


def dfs_iter(successors: NodeNeighbourFn, start: HashableT) -> Iterator[HashableT]:
    visited: set[HashableT] = set()

    def visit(nodes: tuple[HashableT, ...]):
        for n in nodes:
            if n not in visited:
                yield n
                visited.add(n)
                yield from visit(tuple(successors(n)))

    yield from visit((start,))


def bfs_iter_down(
    successors: NodeNeighbourFn, predecessors: NodeNeighbourFn, start: HashableT
) -> Iterator[HashableT]:
    visited: set[HashableT] = set()
    visit_next = sorted(list(successors(start)))
    while len(visit_next) > 0:
        current = visit_next.pop(0)
        if current not in visited:
            visited.add(current)
            yield current
            for child in successors(current):
                if set(predecessors(child)).issubset(visited):
                    visit_next.append(child)


def bfs_iter_up(
    predecessors: NodeNeighbourFn[HashableT],
    successors: NodeNeighbourFn[HashableT],
    start: HashableT,
) -> Iterator[HashableT]:
    return bfs_iter_down(predecessors, successors, start)



import warnings
from typing import Callable, Generic, TypeVar, overload

Tin = TypeVar("Tin")
Tout = TypeVar("Tout")
FN = TypeVar("FN", bound=Callable)


class FunctionDecorator(Generic[FN, Tout]):
    """Apply a callback to functions and either their own or custom names.




    :param: `callback`: will be called as `callback(name, fn)`

    :::{important}
    if you want to use this as a decorator, do not forget to
    return the wrapped function from your callback.
    :::

    *Examples*:
    ```python
    >>> registry = dict()
    >>> def register_fn(name, fn):
    ...  registry[name] = fn
    ...  return fn
    ...
    >>> register = FunctionDecorator(register_fn)
    >>> def my_fn(x):
    ...   print(x)
    ...
    >>> register(my_fn)
    >>> registry["my_fn"]("hello, world!")
    "hello, world!"
    ```

    another example could look like this:

    ```python

    registry = dict()

    def register_fn(name, fn):
        registry[name] = fn
        return fn

    register = FunctionDecorator(register_fn)

    @register("other_name")
    @register
    def my_fn(x):
        print(x)
    ```

    This will add ``my_fn`` to the registry using the
    name ``"my_fn"`` and the name ``"other_name"``.



    """

    def __init__(self, callback: Callable[[str, FN], Tout]):
        self._cb = callback

    @overload
    def __call__(self, name: str, /) -> Callable[[FN], Tout]:
        """Return a function that can be used to register another function with `name`."""
        ...

    @overload
    def __call__(self, fn: FN, /) -> Tout:
        """Register `fn` using its own name."""
        ...

    @overload
    def __call__(self, name: str, fn: FN, /) -> Callable[[FN], Tout]:
        """Register function by name."""
        ...

    def __call__(
        self, arg: FN | str, arg2: FN | None = None, /
    ) -> Tout | Callable[[FN], Tout]:
        if isinstance(arg, str):
            if arg2 is None:
                return self.__reg_by_name(arg)
            else:
                return self.__reg_by_name(arg)(arg2)
        return self.__reg(arg)

    def __reg_by_name(self, name: str) -> Callable[[FN], Tout]:
        def reg(fn: FN) -> Tout:
            return self._cb(name, fn)

        return reg

    def __reg(self, fn: FN) -> Tout:
        return self._cb(fn.__name__, fn)


class FunctionDecoratorDescriptor(Generic[Tin, Tout]):
    """Automatically connect the `FunctionDecorator` to a callback and make it look like a method.

    The owning instance needs to define a callback that has the name `f"_{name}_callback"`,
    where `name` is the name of the field assigned to this descriptor.

    For an example see the `KeyedFunctionDispatcher` below.
    """

    def __set_name__(self, instance, name):
        self._cb_name = f"_{name}_callback"

    def __get__(
        self, instance, owner=None
    ) -> FunctionDecorator[Callable[[Tin], Tout], Callable[[Tin], Tout]]:
        cb = getattr(instance, self._cb_name)

        def wrapped(name: str, fn: Callable[[Tin], Tout]) -> Callable[[Tin], Tout]:
            cb(name, fn)
            return fn

        return FunctionDecorator(wrapped)


class RegisterDescriptor(FunctionDecoratorDescriptor[Tin, Tout]):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        warnings.warn(
            "RegisterDescriptor is deprecated, use FunctionDecoratorDescriptor instead",
            DeprecationWarning,
            stacklevel=2,
        )


class KeyedFunctionDispatcher(Generic[Tin, Tout]):
    """
    Functions can be registered with custom key or by
    using the function's name as key.
    Upon `call` the registry will use the function
    specified by the `dispatch_key_fn` and the argument
    passed to `call` to process the argument and return the result.
    """

    register: FunctionDecoratorDescriptor[Tin, Tout] = FunctionDecoratorDescriptor()

    def __init__(self, dispatch_key_fn: Callable[[Tin], str]) -> None:
        self._key_fn = dispatch_key_fn
        self._fns: dict[str, Callable[[Tin], Tout]] = dict()

    def _register_callback(self, name: str, fn: Callable[[Tin], Tout]) -> None:
        self._fns[name] = fn

    def __contains__(self, item: str | Callable[[Tin], Tout]) -> bool:
        if isinstance(item, str):
            return item in self._fns
        return item.__name__ in self._fns

    def can_dispatch(self, item: Tin) -> bool:
        return self._key_fn(item) in self

    def call(self, arg: Tin) -> Tout:
        if not self.can_dispatch(arg):
            raise ValueError("cannot dispatch function call for {}".format(repr(arg)))
        return self._fns[self._key_fn(arg)](arg)

    def __call__(self, arg: Tin) -> Tout:
        return self.call(arg)



from collections.abc import Iterator
from typing import cast

import torch

from elasticai.creator.nn.design_creator_module import DesignCreatorModule
from elasticai.creator.vhdl.design.design import Design

from .design import Sequential as _SequentialDesign


class Sequential(DesignCreatorModule, torch.nn.Sequential):
    def __init__(self, *submodules: DesignCreatorModule):
        super().__init__(*submodules)

    def create_design(self, name: str) -> Design:
        registry = _Registry()
        submodules = [cast(DesignCreatorModule, m) for m in self.children()]
        for module in submodules:
            registry.register(module.__class__.__name__.lower(), module)
        subdesigns = list(registry.build_designs())
        return _SequentialDesign(
            sub_designs=subdesigns,
            name=name,
        )


class _Registry:
    def __init__(self) -> None:
        self._nodes: dict[str, DesignCreatorModule] = {}
        self._name_counters: dict[str, int] = {}

    def _make_name_unique(self, name: str) -> str:
        return f"{name}_{self._get_counter_for_name(name)}"

    def _get_counter_for_name(self, name: str) -> int:
        if name in self._name_counters:
            return self._name_counters[name]
        else:
            return 0

    def _increment_name_counter(self, name: str):
        self._name_counters[name] = 1 + self._get_counter_for_name(name)

    def register(self, name: str, d: DesignCreatorModule):
        unique_name = self._make_name_unique(name)
        self._nodes[unique_name] = d
        self._increment_name_counter(name)

    def build_designs(self) -> Iterator[Design]:
        for name, module in self._nodes.items():
            yield module.create_design(name)



from collections.abc import Iterable
from dataclasses import dataclass
from unittest import TestCase

from elasticai.creator.file_generation.template import TemplateExpander


@dataclass
class Template:
    content: list[str]
    parameters: dict[str, str | list[str]]


def newline_join(lines: Iterable[str]) -> str:
    return "\n".join(lines)


def expand_template(content: list[str], **parameters: str | list[str]) -> list[str]:
    template = Template(content, parameters)
    return TemplateExpander(template).lines()


def get_result_string(template: list[str], **kwargs: str | list[str]) -> str:
    return newline_join(expand_template(template, **kwargs))


class ExpandTemplatesTestCase(TestCase):
    def test_handles_gracefully_list_of_lines(self) -> None:
        template = ["my", "$value"]
        expected = "my\nvalue"
        actual = get_result_string(template, value=["value"])
        self.assertEqual(expected, actual)

    def test_expand_multiline_with_one_item(self) -> None:
        def check_single_item(value: str):
            template = ["$my_var"]
            expected = value
            values = [value]
            actual = get_result_string(template, my_var=values)
            self.assertEqual(expected, actual)

        for value in ["value", "some other value"]:
            with self.subTest():
                check_single_item(value)

    def test_single_value_with_different_key(self) -> None:
        template = "$my_key".splitlines()
        expected = "value"
        actual = get_result_string(template, my_key=[expected])
        self.assertEqual(expected, actual)

    def test_expand_two_keys(self) -> None:
        template = "$first\n$second".splitlines()
        expected = "ab\ncd"
        actual = get_result_string(template, first=["ab"], second=["cd"])
        self.assertEqual(expected, actual)

    def test_two_values(self) -> None:
        template = "$key".splitlines()
        expected = "a\nb"
        actual = get_result_string(template, key=["a", "b"])
        self.assertEqual(expected, actual)

    def test_no_values(self) -> None:
        template = "$key".splitlines()
        expected = ""
        actual = get_result_string(template, key=[])
        self.assertEqual(expected, actual)

    def test_keep_indentation(self) -> None:
        template = "\t  $key".splitlines()
        expected = "\t  a\n\t  b\n\t  c"
        actual = get_result_string(template, key=["a", "b", "c"])
        self.assertEqual(expected, actual)

    def test_no_error_if_not_all_keys_are_filled(self) -> None:
        template = "$key\n$other_key".splitlines()
        target = ["$key"]
        actual = expand_template(template, other_key=[])
        self.assertEqual(target, actual)

    def test_two_keys_on_one_line(self) -> None:
        template = "\t$first $second".splitlines()
        values = ["0", "1"]
        expected = "\t0 $second\n\t1 $second"
        actual = get_result_string(template, first=values, second=values)
        self.assertEqual(expected, actual)

    def test_expand_two_keys_by_running_twice(self) -> None:
        template = "\t$first $second".splitlines()
        values = ["0", "1"]
        expected = "\t0 0\n\t0 1\n\t1 0\n\t1 1"
        actual = expand_template(template, first=values)
        actual = expand_template(actual, second=values)
        self.assertEqual(expected, newline_join(actual))

    def test_expand_single_string_template_with_single_key(self) -> None:
        template = "$some_key".splitlines()
        expected = "42"
        actual = get_result_string(template, some_key="42")
        self.assertEqual(expected, actual)

    def test_expand_multiple_strings_template(self) -> None:
        template = ["$val1", "$val2", "$val3"]
        expected = "hello\nworld\n42"
        actual = get_result_string(template, val1="hello", val2="world", val3="42")
        self.assertEqual(expected, actual)

    def test_expand_multiple_strings_template_with_multiple_keys_per_line(self) -> None:
        template = ["$val1 $val1 $val1", "$val1 $val2", ""]
        expected = "hello hello hello\nhello world\n"
        actual = get_result_string(template, val1="hello", val2="world")
        self.assertEqual(expected, actual)

    def test_raises_exception_when_inserting_non_existent_keys(self) -> None:
        template = ["$var1", "$var2"]
        with self.assertRaises(KeyError):
            expand_template(template, var3="fail")

    def test_variables_not_filled_returns_empty_set(self) -> None:
        template = Template(["$a", "$b", "$c"], parameters=dict(a="1", b="2", c="3"))
        expander = TemplateExpander(template)
        self.assertEqual(set(), expander.unfilled_variables())

    def test_variables_not_filled_returns_non_empty_set(self) -> None:
        template = Template(["$a", "$b", "$c"], parameters=dict(a="1", c="3"))
        expander = TemplateExpander(template)
        self.assertEqual({"b"}, expander.unfilled_variables())
